# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 20:51:22 2019

@author: yoelr
"""
from .. import Unit
from ._splitter import Splitter
from numpy import ceil

class VibratingScreen(Unit):
    # Assume 3-deck vibrating screen
    _units = {'Area': 'ft^2'}
    _kwargs = Splitter._kwargs
    _run = Splitter._run
    
    #: Flow rate per area of screen per apareture (kg∕ft2-hr-mm)
    capacity = 0.2*907.18474
    
    #: Mesh opening (mm)
    mesh_opening = 8
    
    #: Maximum area of screen (ft^2)
    max_area = 200
    
    def _cost(self):
        results = self._results
        Design = results['Design']
        
        Cost = results['Cost']
        Area = self.ins[0].massnet/(self.capacity*self.mesh_opening)
        Design['N_screens'] = N_screens =  ceil(Area/self.max_area)
        Area /= N_screens
        Design['Area'] = Area
        Cost['Vibrating screens'] = N_screens * self.CEPCI/567 * 1010 * Area **0.91
    


# -*- coding: utf-8 -*-
"""
As BioSTEAM objects are created, they are automatically registered. The `find` object allows the user to find any Unit, Stream or System instance.  When `find` is called, it simply looks up the item and returns it. 

:doc:`Find unit operations and manage flowsheets` 

"""
from graphviz import Digraph
from IPython import display

__all__ = ('find', 'flowsheet_connector', 'Flowsheet')


def make_digraph(units, streams):
    """Return digraph of units and streams."""
    # Create a digraph and set direction left to right
    f = Digraph(format='svg')
    f.attr(rankdir='LR')
    # Set up unit nodes
    U = {}  # Contains units by ID
    UD = {}  # Contains full description (ID and line) by ID
    for u in units:
        u._link_streams()
        graphics = u._graphics
        if not graphics.in_system:
            continue  # Ignore Unit
        
        # Initialize graphics and make Unit node with attributes
        Type = graphics.node_function(u) or u.line
        name = u.ID + '\n' + Type
        f.attr('node', **u._graphics.node)
        f.node(name)
        U[u.ID] = u
        UD[u.ID] = name
        
    keys = UD.keys()

    # Set attributes for graph and streams
    f.attr('node', shape='rarrow', fillcolor='#79dae8',
           style='filled', orientation='0', width='0.6',
           height='0.6', color='black', peripheries='1')
    f.attr('graph', splines='normal', overlap='orthoyx',
           outputorder='edgesfirst', nodesep='0.15', maxiter='1000000')
    f.attr('edge', dir='foward')

    connections = set()
    for s in streams:
        if s.ID == 'Missing Stream':
            continue  # Ignore stream
        
        oU = s._source
        if oU:
            try: oi = oU._outs.index(s) 
            except:
                unit_connection = (oU, s)
                if unit_connection not in connections:
                    oi = oU._outs.index(s._upstream_connection) # The stream says it's source is that unit, but it really means that the unit is the source of the connection stream
                    connections.add(unit_connection)
            oU = oU._ID
        else: oi = None
        
        dU = s._sink
        if dU:
            try:
                di = dU._ins.index(s)
                dU = dU._ID
            except: # The stream is the upstream connection stream
                unit_connection = (dU, s)
                if unit_connection not in connections:
                    di = dU._ins.index(s._downstream_connection) # The stream says it's source is that unit, but it really means that the unit is the source of the connection stream
                    connections.add(unit_connection)
        else: di = None

        # Make stream nodes / unit-stream edges / unit-unit edges
        if oU not in keys and dU not in keys: pass
            # Stream is not attached to anything
        elif oU not in keys:
            # Feed stream case
            f.node(s.ID)
            edge_in = U[dU]._graphics.edge_in
            f.attr('edge', arrowtail='none', arrowhead='none',
                   tailport='e', **edge_in[di])
            f.edge(s.ID, UD[dU])
        elif dU not in keys:
            # Product stream case
            f.node(s.ID)
            edge_out = U[oU]._graphics.edge_out
            f.attr('edge', arrowtail='none', arrowhead='none',
                   headport='w', **edge_out[oi])
            f.edge(UD[oU], s.ID)
        else:
            # Process stream case
            edge_in = U[dU]._graphics.edge_in
            edge_out = U[oU]._graphics.edge_out
            f.attr('edge', arrowtail='none', arrowhead='normal',
                   **edge_in[di], **edge_out[oi])
            f.edge(UD[oU], UD[dU], label=s.ID)
    return f


# %% Flowsheet search      
    
class Flowsheet:
    """Create a Flowsheet object which stores references to all stream, unit, and system objects."""
    #: [dict] All flowsheets
    flowsheet = {}
    
    def __init__(self, ID):
        #: [str] ID of flowsheet
        self.ID = ID
        
        #: [dict] Dictionary of systems
        self.system = {}
        
        #: [dict] Dictionary of units
        self.unit = {}
        
        #: [dict] Dictionary of streams
        self.stream = {}
        
        self.flowsheet[ID] = self
    
    def diagram(self, kind='surface'):
        """Display all units and attached streams.
        
        **kind:** Must be one of the following:
                * **'thorough':** Thoroughly display every unit.
                * **'surface':** Display units and recycle systems.
                * **'minimal':** Minimally display flowsheet as a box with feeds and products.
        
        """
        if kind == 'thorough':
            return self._thorough_diagram()
        elif kind == 'surface':
            return self._surface_diagram()
        elif kind == 'minimal':
            return self._minimal_diagram()
        else:
            raise ValueError(f"kind must be either 'thorough', 'surface', or 'minimal'.")
    
    def _thorough_diagram(self):
        units = list(self.unit.values())
        units.reverse()
        streams = set()
        for u in units:
            streams.update(u._ins)
            streams.update(u._outs)
        streams.difference_update(self._upstream_connections)
        x = display.SVG(make_digraph(units, streams).pipe(format='svg'))
        display.display(x)
        # img = make_digraph(units).pipe('png')
        # display.display(display.Image(img))
    
    def _minimal_diagram(self):
        from . import system, Stream
        streams = self.stream.values()
        feeds = set(filter(system._isfeed, streams))
        products = set(filter(system._isproduct, streams))
        product = Stream(None)
        product._ID = ''
        feed = Stream(None)
        feed._ID = ''
        system._streamUnit('\n'.join([i.ID for i in feeds]),
                           feed)
        system._streamUnit('\n'.join([i.ID for i in products]),
                           None, product)
        unit = system._systemUnit(self.ID, product, feed)
        unit.line = 'flowsheet'
        unit.diagram(1)
        
    def _surface_diagram(self):
        from . import system, Stream
        units = set(self.unit.values())
        StrUnit = system._streamUnit
        refresh_units = set()
        for i in self.system.values():
            if i.recycle and not any(sub.recycle for sub in i.subsystems):
                outs = []
                ins = []
                feeds = []
                products = []
                for s in i.streams:
                    source = s._source
                    sink = s._sink
                    if source in i.units and sink not in i.units:
                        if sink: outs.append(s)
                        else: products.append(s)
                        refresh_units.add(source)
                    elif sink in i.units and source not in i.units:
                        if source: ins.append(s)
                        else: feeds.append(s)
                        refresh_units.add(sink)
                
                if len(feeds) > 1:
                    feed = Stream(None)
                    feed._ID = ''
                    units.add(StrUnit('\n'.join([i.ID for i in feeds]), feed))
                    ins.append(feed)
                else: ins += feeds
                
                if len(products) > 1:
                    product = Stream(None)
                    product._ID = ''
                    units.add(StrUnit('\n'.join([i.ID for i in products]),
                                      None, product))
                    outs.append(product)
                else: outs += products
                
                subsystem_unit = system._systemUnit(i.ID, outs, ins)
                units.difference_update(i.units)
                units.add(subsystem_unit)
        
        sys = system.System(None, units)
        sys._thorough_diagram()
        for i in refresh_units:
            i.ins = i._ins
            i.outs = i._outs
    
    def __call__(self, item_ID) -> 'item':
        """Return requested biosteam item.
    
        **Parameters**
    
             **item_ID:** [str] ID of the requested item.
    
        """
        item_ID = item_ID.replace(' ', '_')
        obj = (self.stream.get(item_ID)
               or self.unit.get(item_ID)
               or self.system.get(item_ID))
        if not obj: raise ValueError(f"No registered item '{item_ID}'")
        return obj
    
    def __repr__(self):
        return f'<{type(self).__name__}: {self.ID}>'


class find(Flowsheet):
    """Create a find object which can search through flowsheets."""
    __slots__ = ()
    _upstream_connections = set()
    
    @property
    def mainflowsheet(self):
        """[Flowsheet] Main flowsheet that is updated with new biosteam objects"""
        return find._mainflowsheet
    
    @mainflowsheet.setter
    def mainflowsheet(self, flowsheet):
        if isinstance(flowsheet, Flowsheet):
            find.__dict__ = flowsheet.__dict__
        else:
            raise TypeError('Main flowsheet must be a Flowsheet object')
        find._mainflowsheet = flowsheet
    
    def __new__(cls):
        raise TypeError('Cannot create new Find object.')

    def __repr__(self):
        return f'<{type(self).__name__}: mainflowsheet={self.ID}>'
    
    
#: [find] Find BioSTEAM objects by ID.
find = object.__new__(find)
find.mainflowsheet = Flowsheet('Default')


# %% Connect between different flowsheets

def flowsheet_connector(upstream, downstream, species=None):
    """Return a function that copies specifications from `upstream` to `downstream`. This serves to connect different flowsheets.
    
    **Parameters**
    
        **upstream:** [Stream] Stream that will be copied to `downstream`.
        
        **downstream:** [Stream] Flow rate, T, P, and phase information will be copied from `upstream` to this stream.
        
        **species:** list[str] IDs of species to be passed down. If None, all species in common will be passed.
    
    """
    # Source and sink.
    upstream._sink = downstream._sink
    downstream._source = upstream._source
    downstream._upstream_connection = upstream
    upstream._downstream_connection = downstream
    find._upstream_connections.add(upstream)
    if not species:
        species = set(upstream._IDs).intersection(downstream._IDs)
    upindex = upstream.indices(*species)
    downindex = downstream.indices(*species)
    def connect():
        # Flow rate, T, P and phase
        downstream._molarray[downindex] = upstream._molarray[upindex]
        downstream.T = upstream.T
        downstream.P = upstream.P
        downstream.phase = upstream.phase
    return connect


        
    # #: [bool] True if flowsheet is contantly rendered
    # self.live = True

# def _show_once(self):
    #     fig = plt.figure()
    #     plt.axis('off')
    #     f = open('diagram.png', 'wb')
    #     diagram = make_digraph(self.unit.values()).pipe(format='png')
    #     f.write(diagram)
    #     img = mpimg.imread('diagram.png')
    #     plt.imshow(img, interpolation='spline36')
    #     fig.show()
    #     # img = Image.open('diagram.png')
    #     # img.show() 
    
    # def _show(self):
    #     fig = plt.figure()
    #     plt.axis('off')
    #     f = open('diagram.png', 'wb')
    #     for i in range(3):
    #         diagram = make_digraph(self.unit.values()).pipe(format='png')
    #         f.write(diagram)
    #         img = mpimg.imread('diagram.png')
    #         plt.clf() 
    #         plt.imshow(img, interpolation='spline36')
    #         fig.show()
    #         time.sleep(10)
    #     f.close()
    #     os.remove('diagram.png')
    
    # def show(self):
    #     t1 = threading.Thread(target=self._show())
    #     t1.start()


# -*- coding: utf-8 -*-
"""
Created on Wed Dec  5 16:42:49 2018

This module includes classes and functions relating arrays.

@author: Guest Group
"""
import numpy as np
from array_collections import array, tuple_array, material_array, \
                              property_array, same_type
from free_properties import PropertyFactory

__all__ = ('array', 'tuple_array', 'material_array', 'property_array',
           'PropertyFactory', 'same_type', 'reorder', 'fraction')

ndarray = np.ndarray
asarray = np.asarray

material_array._units = 'kmol/hr'

# %% Functions

def reorder(array: 'array_like', current_order: tuple, new_order: tuple) -> np.ndarray:
    """Return a reordered array with zero in place of elements that exist in the new order but not in the current order.
    
    .. code-block:: python
       
       >>> reorder([1,2,3], ('a', 'b', 'c'), ('a', 'c', 'd', 'b'))
       array([1, 3, 0, 2])
    
    """
    n = 0
    p_new = np.zeros(len(new_order), dtype=float)
    for ID1 in current_order:
        m = 0
        for ID2 in new_order:
            if ID1 is ID2:
                p_new[m] = array[n]
            m += 1
        n += 1
    return p_new

def fraction(array) -> 'Normalized array':
    """Normalize array to a magnitude of 1. If magnitude is zero, all fractions will have equal value."""
    sum_array = array.sum()
    if sum_array == 0:
        l = len(array)
        frac = np.ones(l)/l
    else:
        frac = array/sum_array
    return frac


# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 11:10:49 2019

@author: yoelr
"""
from .. import Unit
import numpy as np

class ConveyingBelt(Unit):
    length = 50 #: ft
    height = 30 #: ft
    _N_outs = 1
    _has_power_utility = True
    _has_proxystream = True
    _bounds = {'Volumetric flow': (120, 500)}
    
    def _run(self):
        self.outs[0].copylike(self.ins[0])
    
    def _cost(self):
        feed = self.ins[0]
        r = self._results
        Design = r['Design']
        volbounds = self._bounds['Volumetric flow']
        volnet = feed.volnet*35.315 # ft3/hr
        N_belts = volnet/volbounds[-1]
        Design['N_belts'] = N_belts = np.ceil(N_belts)
        volnet /= N_belts 
        massnet = feed.massnet*0.0006124/N_belts #lb/s
        power = N_belts * 0.00058*massnet**0.82*self.length + self.height*0.00182*massnet# hp
        power *= 0.7457 # kW
        self._power_utility(power)
        r['Cost']['Conveying belt and motor'] = N_belts * self.CEPCI/567 * 813*volnet**0.38
        




# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 14:36:58 2018

@author: yoelr
"""
import copy

__all__ = ('Graphics',)


# %%

_edge_in = [{'headport': 'c'} for i in range(3)]
_edge_out = [{'tailport': 'c'} for i in range(2)]
_node = {'shape': 'box',
         'fillcolor': "white:#CDCDCD",
         'style': 'filled',
         'gradientangle': '0',
         'width': '0.6',
         'height': '0.6',
         'orientation': '0.0',
         'color': 'black',
         'peripheries': '1',
         'margin': 'default'}

# %%
    
class Graphics:
    """Create a Graphics object that contains specifications for Graphviz node and edge styles"""
    in_system = True #: [bool] True for Unit object to appear within a system diagram
    def __init__(self, edge_in=_edge_in,
                 edge_out=_edge_out,
                 node=_node,
                 node_function=lambda unit: None,
                 name=None):
        self.edge_in = copy.deepcopy(edge_in)
        self.edge_out = copy.deepcopy(edge_out)
        self.node = copy.deepcopy(node)
        self.node_function = node_function
        self.name = None

default_graphics = Graphics()



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 21:23:56 2018

@author: yoelr
"""
from .. import Unit, MixedStream
from ._flash import RatioFlash, PartitionFlash
from ._splitter import Splitter
from copy import copy

__all__ = ('Centrifuge_LLE', 'RatioCentrifuge_LLE',
           'SplitCentrifuge_LLE', 'PartitionCentrifuge_LLE')

class Centrifuge_LLE(Unit):
    r"""Create an equlibrium based centrifuge with the option of having liquid non-keys and LIQUID non-keys completly separate into their respective phases.

    Equation for f.o.b cost:

    :math:`C_{fob}^{2007} = 28100 Q^{0.574} (0.1 < Q < 100 \frac{m^3}{h})` 

    **Parameters**

        **species_IDs:** *tuple[str]* IDs of equilibrium species
    
        **split:** *tuple[float]* Initial guess split fractions of equilibrium specise to the 'liquid' phase.
    
        **lNK:** *tuple[str]* Species assumed to completely remain in the 'liquid' phase.
    
        **LNK:** *tuple[str]* Species assumed to completely remain in the 'LIQUID' phase.
    
        **solvents:** *tuple[str]* Species corresponding to specified solvent_split.
    
        **solvent_split:** *tuple[float]* Split fractions of each specie to the 'liquid' phase.
         
    **ins**
    
        [0] Input stream
        
    **outs**
    
        [0] 'liquid' phase stream
        
        [1] 'LIQUID' phase stream
    
    **Examples**
    
        :doc:`Centrifuge_LLE Example`
    
    """
    line = 'Liquids Centrifuge'
    
    _kwargs = {'species_IDs': None,
               'split': None,
               'lNK': (),
               'LNK': (),
               'solvents': (),
               'solvent_split': ()}

    _bounds = {'Flow rate': (0.1, 100)}
    _units = {'Flow rate': 'm^3/hr'}
    electricity_rate = 3.66 #: kW/(m3/hr) from USDA biosdiesel Super Pro model
    # Possibly 1.4  kW/(m3/hr)
    # https://www.sciencedirect.com/topics/engineering/disc-stack-centrifuge
    # Microalgal fatty acids—From harvesting until extraction H.M. Amaro, ... A. Catarina Guedes, in Microalgae-Based Biofuels and Bioproducts, 2017
    _has_power_utility = True

    def _setup(self):
        liq, LIQ = self.outs
        liq.phase = 'l'
        LIQ.phase = 'l'
        self._cached = cached = {}
        cached['mixed stream'] =  MixedStream(None)
        if self._kwargs['species_IDs'] is None:
            self._kwargs['species_IDs'] = liq.species_IDs

    def _run(self):
        liq, LIQ = self.outs
        feed = self.ins[0]

        kwargs = self._kwargs
        LLE_kwargs = copy(kwargs)

        ms = self._cached['mixed stream']
        ms.empty()
        ms.liquid_mol = feed.mol
        ms.LLE(T=feed.T,  **LLE_kwargs)
        liq.mol = ms.liquid_mol
        LIQ.mol = ms.LIQUID_mol
        liq.T = LIQ.T = ms.T
        liq.P = LIQ.P = ms.P

    def _design(self):
        Design = self._results['Design']
        Design['Flow rate'] = self._volnet_out
        return Design

    def _cost(self):
        r = self._results
        Design = r['Design']
        Cost = r['Cost']
        Q = Design['Flow rate']
        self._power_utility(self.electricity_rate*Q)
        Cost['Centrifuge'] = 28100 * Q ** 0.574 * self.CEPCI/525.4
        return Cost

class PartitionCentrifuge_LLE(Centrifuge_LLE):
    _N_heat_utilities = 0
    kwargs = PartitionFlash._kwargs
    _run = PartitionFlash._run 
    _setup = PartitionFlash._setup
    def _set_phases(self):
        top, bot = self.outs
        top.phase = 'l'
        bot.phase = 'L'

class RatioCentrifuge_LLE(Centrifuge_LLE):
    _N_heat_utilities = 0
    _kwargs = {'Kspecies': [],  # list of species that correspond to Ks
               'Ks': [],  # list of molar ratio partition coefficinets,
               # Ks = y/x, where y and x are molar ratios of two different phases
               'top_solvents': [],  # list of species that correspond to top_split
               'top_split': [],  # list of splits for top_solvents
               'bot_solvents': [],  # list of species that correspond to bot_split
               'bot_split': []}  # list of splits for bot_solvents
    def _setup(self): pass
    _run = RatioFlash._run
    
class SplitCentrifuge_LLE(Centrifuge_LLE):
    _kwargs = Splitter._kwargs
    _setup = Splitter._setup
    _run = Splitter._run


# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 11:32:01 2019

@author: yoelr
"""

from .. import Unit

class MagneticSeparator(Unit):
    _N_outs = 1
    _has_proxystream = True
    
    #: Original CEPCI
    CEPCI_0 = 576
    
    #: Original Price (USD)
    C_0 = 533571.0
    
    #: Original flow rate (kg/hr)
    V_0 = 333333
    
    #: Scaling exponent
    exp = 0.6
    
    def _run(self):
        self.outs[0].copylike(self.ins[0])
    
    def _cost(self):
        massflow = self.ins[0].massnet
        self._results['Cost']['Magnetic separator'] = self.CEPCI/self.CEPCI_0*self.C_0 * (massflow/self.V_0)**self.exp


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:07:01 2018

@author: yoelr
"""
from .. import Unit
from ._hx import HXutility

class EnzymeTreatment(Unit):
    _N_outs = 1
    _N_heat_utilities = 1
    _has_power_utility = True
    _kwargs = {'T': 298.15}  # operating temperature (K)
    
    #: Fraction of filled tank to total tank volume
    V_wf = 0.8
    
    #: Residence time (hr)
    tau = 1
    
    #: Electricity rate (kW/m3)
    electricity_rate = 0.591
    
    def _init(self):
        self._heat_exchanger = he = HXutility(None, None) 
        self._heat_utilities = he._heat_utilities
        he._ins = self._ins
        he._outs = self._outs
        
    def _setup(self):
        T = self._kwargs['T']
        self.outs[0].T = T
    
    def _run(self):
        feed = self.ins[0]
        out = self.outs[0]
        out.mol = self._mol_in
        out.phase = feed.phase
        out.P = feed.P
    
    def _cost(self):
        he = self._heat_exchanger
        he._design()
        he._cost()
        
        r = self._results
        Cost = r['Cost']
        Cost['Heat exchanger'] = he._results['Cost']['Heat exchanger']
        V = self.tau * self.ins[0].volnet / self.V_wf
        self._power_utility(self.electricity_rate*V)
        Cost['Tank'] = 12080 * V**0.525 * self.CEPCI/525.4 # Same equation as MixTank
    


# -*- coding: utf-8 -*-
"""
Created on Wed Apr 10 11:48:29 2019

@author: yoelr
"""

from ._unit import metaUnit

class metaFinal(metaUnit):
    """Make Unit class final/unable to be inherited."""
    def __new__(mcl, name, bases, dct):
        # Make this class a final class
        for b in bases:
            if isinstance(b, mcl):
                raise TypeError(f"Cannot inherit from {b}. Instances of {mcl.__name__} cannot be inherited.")

        return super().__new__(mcl, name, bases, dct)


# -*- coding: utf-8 -*-
"""
Created on Thu Mar  7 23:17:38 2019

@author: yoelr
"""
import numpy as np
import pandas as pd
from .. import Unit
from ._block import Block

__all__ = ('Grid',)

DF = pd.DataFrame

# %% Functions

def _blockunit(blockfunc_args):
    element = blockfunc_args[0]._block._element
    if isinstance(element, Unit): return element
    else: return element._sink

def _fill_space(args, num_stack, argspace, thread, tuple_):
    for i, f, fargs in num_stack:
        for a in fargs[1:]:
            args[i] = a
            argspace.append(tuple_(args))
            thread.append((f, a))
            _fill_space(args, num_stack[:i], argspace, thread, tuple_)
        fargs.reverse()

def _ordered_space(args, num_args, argspace, tuple_, space):
    for i, fargs in num_args:
        for a in fargs[1:]:
            args[i] = a
            new = tuple_(args)
            if new in space:
                argspace.append(new)
            _ordered_space(args, num_args[:i], argspace, tuple_, space)
        fargs.reverse()

# %% Grid of simulation blocks

class Grid:
    """Create a Grid object that allows for optimized evaluation over an argument space.
    
    **Parameters**
    
        **system:** [System] Metric parameters should act on the system.
    
        **metric:** [function] Should return metric valye.
        
        **args:** [tuple] Arguments to pass to the `metric` function.
        
        **ID:** [str] ID of the Grid object
        
        **layout:** {'simple', 'cartesian'}
            * 'simple': The argument space is a matrix of the parameter values.
            * 'cartesian': The argument space is a cartesian product of the parameter values.
    
    **Examples**

         :doc:`Grid Example`
    
    """
    __slots__ = ('_table',
                 '_system',
                 '_stack',
                 '_metric',
                 '_ID',
                 '_layout')
    
    def __init__(self, system, metric, *, args=(), ID=None, layout='simple'):
        if args:
            metric_no_args = lambda: metric(*args)
            metric_no_args.__name__ = metric.__name__
            self._metric = metric_no_args
        else: self._metric = metric
        self.layout = layout
        self._system = system
        self._ID = ID or self._metric.__name__
        
        #: [(function, args)] Iterable of block fuctions and respective arguments.
        self._stack = []
        
        #: [DataFrame] Table of the argument space with results in the final column.
        self._table = None
    
    @property
    def layout(self):
        """
        Must be either 'simple' or 'cartesian'
            * 'simple': The argument space is a matrix of the parameter values.
            * 'cartesian': The argument space is the cartesian product of parameter values.
        """
        return self._layout
    
    @layout.setter
    def layout(self, layout):
        layout = layout.casefold()
        if layout not in ('simple', 'cartesian'):
            if isinstance(layout, str): layout = f"'{layout}'"
            raise ValueError(f"Grid layout must be either 'simple' or 'cartesian', not {layout}")
        self._layout = layout
    
    @property
    def table(self):
        """[DataFrame] Table of the argument space with results in the final column."""
        if not self._stack: return None
        elif self._table is None: self._loadtable()
        return self._table
    
    def evalparam(self, element, setter, values, isolated=False):
        """Return metric at given parameter values.
        
        **Parameters**
        
            **element:** [Unit or Stream] Element in the system being altered.
            
            **setter:** [function] Should set parameter in the element.
            
            **values:** [iterable] Values for parameter.
            
            **isolated** [bool] If True, account for downstream operations. If False, only account for element.
        
        """
        system = None if isolated else self._system
        blockfunc = Block(element, system)(self._metric, setter)
        return [blockfunc(i) for i in values]
    
    def addparam(self, element, setter, values, isolated=False):
        """Add parameter to vary in metric.
        
        **Parameters**
        
            **element:** [Unit or Stream] Element in the system being altered.
            
            **setter:** [function] Should set parameter in the element.
            
            **values:** [iterable] Values for parameter.
            
            **isolated** [bool] If True, account for downstream operations. If False, only account for element.
        
        """
        system = None if isolated else self._system
        self._stack.append((Block(element, system)(self._metric, setter), values))
    
    def _loadtable(self):
        """Load argument space and return parameters for simulation.
        
        **Returns**
        
            **argspace:** [array] All arguments
            **funcs:** [list] All block functions
            **initial_args:** [tuple] First arguments to be simulated
            **thread:** [list] All block functions and arguments to be simulated.
        
        """
        system = self._system
        stack = self._stack
        if not stack: raise RuntimeError(f'No metric parameters set for {repr(self)} object.')
        tuple_ = tuple
        # Order blocktests from last in network to first
        length = len(system._unitnetwork)
        index = system._unitnetwork.index
        stack.sort(key=lambda x: index(_blockunit(x))
                                 if x[0]._block._system else length,
                   reverse=True)
        
        initial_args = []
        funcs = []
        element_names = []
        all_args = []
        list_ = list
        if self._layout == 'cartesian':
            for func, args in stack:
                initial_args.append(args[0])
                funcs.append(func)
                element = func._block._element
                element_names.append(element.line + '-' + element.ID.replace('_', ' '))
                all_args.append(list_(args))
            num_stack = tuple_(zip(range(len(funcs)), funcs, all_args))
            args = initial_args
            initial_args = tuple_(initial_args)
            argspace = [initial_args]
            thread = [(funcs[0], initial_args[0])]
            _fill_space(args, num_stack, argspace, thread, tuple_)
        else:
            argspace = []
            for func, args in stack:
                argspace.append(args)
                funcs.append(func)
                element = func._block._element
                element_names.append(element.line + '-' + element.ID.replace('_', ' '))
            argspace = list_(zip(*argspace))
            key = lambda x: x[i]
            index = list_(range(len(stack)))
            index.reverse()
            for i in index: argspace.sort(key=key)
            args = initial_args = argspace[0]
            thread = None
        
        paramIDs = [i._param.capitalize().replace('_', ' ')
                    for i in funcs]
        spacelen = len(argspace)
        spacerange = range(spacelen)
        
        self._table = DF(argspace,
                         columns=pd.MultiIndex.from_arrays((element_names, paramIDs),
                                                           names=('Element', 'Parameter')),
                         index=spacerange)
        self._table[self._ID] = (None,)*spacelen
        
        return (np.asarray(argspace),
                funcs,
                initial_args,
                thread)
        
    def simulate(self):
        """Simulate Grid object over the argument space and save metric values to `table`."""
        # Setup units before simulation
        argspace, funcs, initial_args, thread = self._loadtable()
        for func, arg in zip(funcs, initial_args):
            func._setter(arg)
        
        # Simulate system to initialize whole network
        try: self._system.simulate()
        except: pass
        values = []
        add = values.append
        if thread:
            for func, arg in thread:
                try: add(func(arg))
                except: add(None)
        else:
            metric = self._metric
            add(metric())
            last_args = initial_args
            simulate_funcs = [i._block._simulate for i in funcs]
            setter_funcs = [i._setter for i in funcs]
            index = tuple(range(len(last_args)))
            for args in argspace[1:]:
                simfunc_pos = None
                for i, j, k in zip(index, args, last_args):
                    if j != k:
                        setter_funcs[i](j)
                        simfunc_pos = i
                if simfunc_pos: simulate_funcs[simfunc_pos]()
                add(metric())
                last_args = args
                
        self._table[self._ID] = values
        
    def _repr(self):    
        return f'{type(self).__name__}: {self._ID}'
    
    def __repr__(self):
        return f'<{self._repr()}>'
       
    def _info(self):
        if not self._stack:
            return (f'{self._repr()}\n'
                    +' element  parameters\n'
                    +' None     None')
        blocks = {}
        for i, _ in self._stack:
            e = i._block._element
            p = i._param
            blk = f'{type(e).__name__}-' + f'{e}'.replace('_', ' ')
            if blk not in blocks: blocks[blk] = [p]
            else: blocks[blk].append(p)
        
        lenghts_block = []
        lines = []
        for blk, params in blocks.items():
            blklen = len(blk)
            blklen_spaces = (blklen)*' '
            newlines = []
            newlines.append(f" {blk}${params[0].replace('_', ' ')}\n")
            for p in params[1:]:
                newlines.append(f" {blklen_spaces}${p.replace('_', ' ')}\n")
            lines.extend(newlines)
            lenghts_block.extend((blklen,)*len(newlines))
        
        maxlen_block = max(lenghts_block)
        out = f'{self._repr()}\n'
        maxlen_block = max(maxlen_block, 7)
        out += ' element' + (maxlen_block - 7)*' ' + '  parameter\n'
        for newline, len_ in zip(lines, lenghts_block):
            newline = newline.replace('$', ' '*(maxlen_block-len_) + '  ')
            out += newline
        
        return out.rstrip('\n ')
    
    def show(self):
        """Return information on metric parameters."""
        print(self._info())
        
        
        


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:18:16 2018

@author: yoelr
"""

from .._unit import Unit
from ..exceptions import DesignError
from . import Splitter
import numpy as np

def checkbounds(x, bounds):
    lb, up = bounds
    return lb < x < up

class Clarifier(Unit):
    _has_power_utility = True
    _kwargs = Splitter._kwargs
    _run = Splitter._run
    
    # Height of the clarifier tank from other designs, estimate (ft)
    height = 10
    
    # Setting the working bounds for different materials
    _bounds = {'Area steel' : (80, 8000), 'Area concrete': (8000, 125000)}
    _units =  {'Settling area': 'ft^2'}
    
    
    def _cost(self):
        results = self._results
        Design = results['Design']
        overflow = self.ins[0]
        
        # Heuristic settling area estimation
        # Settling area in ft^2 = overflow in gpm
        SetArea = overflow.volnet *  4.4028
        Design['Settling area'] = SetArea
        power = self._calc_energy(SetArea)*0.7457 # in kW
        self._power_utility(power) 
        
        # Checking to see which cost equation/material to use
        Steel_bounds, Concrete_bounds = self._bounds.values()
        if checkbounds(SetArea, Steel_bounds):
            Cost = 3810*SetArea**0.58
            Design['Material'] = 'Steel'
        elif checkbounds(SetArea, Concrete_bounds):
            Cost = 2720*SetArea**0.58
            Design['Material'] = 'Concrete'
        else:
            raise DesignError('Volumetric flow rate is out of working range')
        results['Cost']['Clarifier'] = Cost*self.CEPCI/567
        
    def _calc_energy(self, SetArea):
        # Finding diameter of the tank in ft
        diameter = np.sqrt(SetArea*4/np.pi)
        # Energy assuming quadratic relationship in hp
        Energy = 16 * (diameter/200)**2
        return Energy


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:56:55 2018

@author: yoelr
"""
import numpy as np
from .. import Unit
from . import Tank


class Reactor(Unit):
    """Abstract class for reactors."""
    

# %% Reactor classes
    
class BatchReactor(Reactor):
    
    _N_heat_utilities = 1
    
    # Exponential scaling
    C_0 = None #: Original Price
    V_0 = None #: Original Volume
    exp = None #: Scaling exponent
    
    @classmethod
    def _solve(cls, 
               v_0: 'Flow rate',
               tau: 'Reaction time',
               tau_0: 'Cleaning and unloading time',
               N_reactors: 'Loading time per volume',
               V_wf: 'Fraction of working volume') -> 'Results [dict]':
        
        V_T = v_0*(tau + tau_0)/(1-1/N_reactors) # Reacting volume
        V_i = V_T/N_reactors
        t_L = V_i/v_0
        t_B = tau + tau_0 + t_L
        V_i /= V_wf
        
        return {'Reactor volume': V_i,
                'Cycle time': t_B, 
                'Loading time': t_L}
    
    
# %%    



# -*- coding: utf-8 -*-
"""
Created on Wed Mar  6 16:46:04 2019

@author: yoelr
"""

__all__ = []

from . import vacuum
from . import design_tools
from .vacuum import *
from .design_tools import *

__all__.extend(vacuum.__all__)
__all__.extend(design_tools.__all__)


# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 16:40:50 2019

@author: yoelr
"""
from .. import Unit
import numpy as np

class ScrewFeeder(Unit):
    length = 30 #: ft
    _N_outs = 1
    _has_power_utility = True
    _has_proxystream = True
    
    _bounds = {'Volumetric flow': (400, 10000)}
    
    def _run(self):
        self.outs[0].copy_like(self.ins[0])
    
    def _cost(self):
        feed = self.ins[0]
        r = self.results
        Design = r['Design']
        volbounds = self.bounds['Volumetric flow']
        volnet = feed.volnet*35.315 # ft3/hr
        N = volnet/volbounds[-1]
        Design['N'] = N = np.ceil(N)
        volnet /= N 
        massnet = feed.massnet*0.0006124/N #lb/s
        power = N * 0.0146*massnet**0.85*self.length # hp
        power *= 0.7457 # kW
        self._power_utility(power)
        r['Cost']['Conveying belt and motor'] = N * self.CEPCI/567 * 813*volnet**0.38


# -*- coding: utf-8 -*-
"""
Created on Wed Mar  6 16:53:04 2019

@author: yoelr
"""
from . import _block
from . import _grid
from ._block import *
from ._grid import *

__all__ = []

__all__.extend(_block.__all__)
__all__.extend(_grid.__all__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 16 11:49:52 2018

This module includes classes and functions relating exception handling.

@author: Yoel Rene Cortes-Pena
"""
import sys as _sys

__all__ = ('biosteamError', 'Stop', 'DesignError', 'SolverError', 'EquilibriumError','DimensionError', 'DesignWarning')

# %% Biosteam errors

class biosteamError(Exception):
    """General exception in biosteam."""

class Stop(biosteamError):
    """Exception to Stop code."""

class DesignError(biosteamError):
    """Exception regarding unit design."""

class SolverError(biosteamError):
    """Exception regarding solvers."""

class EquilibriumError(biosteamError):
    """Exception regarding equilibrium."""

class DimensionError(biosteamError):
    """Exception regarding wrong dimensions."""


# Python's built in KeyError quotes the message, used as a by-pass for the debbuger
_KE = type('KeyError', (biosteamError, ), {})


# %% Biosteam Warnings

class DesignWarning(Warning):
    """Warning regarding design constraints."""
    
#%% Decorators and functions

def _notify_error(func):
    """Decorate class method to provide a location summary when an error occurs."""
    if hasattr(func, '_original'):
        func = func._original

    def wrapper(self, *args, **kwargs):
        try: return func(self, *args, **kwargs)
        except Exception as e:
            # If exception already include location, it is replaced
            location = f'@{type(self).__name__} {self}'
            msg = str(e).strip('\n').replace(location + ': ', '')
            
            # Add location to message
            if not ('@' in msg and ':\n' in msg):
                msg = location + f'.{func.__name__}:\n' + msg                 
            
            # Raise exception with same traceback but new message
            if type(e) is KeyError:
                raise _KE(msg).with_traceback(_sys.exc_info()[2])
            else:
                raise type(e)(msg).with_traceback(_sys.exc_info()[2])
    
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    wrapper._original = func
    wrapper.__annotations__ = func.__annotations__
    return wrapper






# -*- coding: utf-8 -*-
"""
Created on Tue Sep 18 10:28:29 2018

@author: Guest Group
"""
from scipy.optimize import least_squares, brentq, minimize_scalar, newton
from ._species import Species
from ._stream import Stream, nonzero_species, MassFlow, \
                            mol_flow_dim, mass_flow_dim, vol_flow_dim
from ._utils import fraction, tuple_array, material_array, \
                           PropertyFactory, property_array
from .exceptions import EquilibriumError, DimensionError
from . import _Q
import numpy as np


ln = np.log
exp = np.exp

__all__ = ('MixedStream',)


# %% Dictionaries for finding the right attribute given the phase

phases = ('s', 'l', 'L', 'g')
range4 = (0, 1, 2, 3)
phase_index = dict(zip(phases, range4))

# Used for the show method
letter_phases = {'s': 'solid',
                 'l': 'liquid',
                 'L': 'LIQUID',
                 'g': 'vapor'}

# Used during MixedStream __init__
molflows_ID = ('_solid_mol',
               '_liquid_mol',
               '_LIQUID_mol',
               '_vapor_mol')

volflows_ID = ('_solid_vol',
               '_liquid_vol',
               '_LIQUID_vol',
               '_vapor_vol')

massflows_ID = ('_solid_mass',
                '_liquid_mass',
                '_LIQUID_mass',
                '_vapor_mass')

phases_volflow = {'s': 'solid_vol',
                  'l': 'liquid_vol',
                  'L': 'LIQUID_vol',
                  'g': 'vapor_vol'}


# %% Volumetric flow property

@PropertyFactory    
def VolumetricFlow(self):
    """Volumetric flow (m^3/hr)."""
    stream, mol, phase = self.data
    if mol:
        c = self.name # c: compound
        c.T = stream.T
        c.P = stream.P
        c.phase = phase
        return (c.Vm * mol.item(0) * 1000)
    else:
        return 0.

@VolumetricFlow.setter
def VolumetricFlow(self, value):
    if value:
        stream, mol, phase = self.data
        c = self.name # c: compound
        c.T = stream.T
        c.P = stream.P
        c.phase = phase
        mol[0] = value/(c.Vm * 1000)
    else:
        mol[0] = 0.


# %% MixedStream (4 phase flow)

# TODO: Add solids handling:
# -Viscosity with class attribute J, the coefficient in Mc = Mw(1 - Vs/Vm)^(1/J).
# A generalized mixture rule for estimating the viscosity of solid-liquid suspensions and mechanical properties of polyphase rocks and composite materials, https://agupubs.onlinelibrary.wiley.com/doi/epdf/10.1029/2004JB003124, suggests that J = -0.5 for real suspensions including carbon fibers

# TODO: Make MixedStream capable of handling equilibrium, and mixing with Stream


class MixedStream(Stream):
    """An extension of the :doc:`Stream` class. All inherited properties and methods are supported. Create a MixedStream object that can manage mixtures of solid ('s'), liquid ('l'), LIQUID ('L'), and vapor ('g') phases. A MixedStream object is capable of vapor-liquid equilibrium (VLE) and liquid-LIQUID equilibrium (LLE).

    **Parameters**

         **ID:** [str] ID of the stream. If ID is '', a default ID will be chosen.

         **species:** [Species] Species object for the stream. If None, assume same as class' Species object.

         **T:** [float] Temperature (K)

         **P:** [float] Pressure (Pa)

    **Examples**

         :doc:`MixedStream objects and thermodynamic equilibrium`
    """
    #: The default phase when flow rates are zero.
    _default_phase = 'l'
    
    #: Unit source
    _source = None
    
    #: Unit sink
    _sink = None
    
    def __init__(self, ID='', species=None, T=298.15, P=101325):
        # Get species and set species information
        if species is None:
            self._copy_species(self)
        elif isinstance(species, Species):
            self._set_species(species)
        else:
            raise TypeError(f"species must be a Species object, not '{type(species).__name__}'")
        
        self.ID = ID
        
        #: [float] Temperature (K)
        self.T = T
        
        #: [float] Pressure (Pa)
        self.P = P
        self._dew_cached = {}
        self._bubble_cached = {}
        self._y_cached = {}
        self._lL_split_cached = {}
        self._setflows(np.zeros((4, self._Nspecies), dtype=np.float64))
        

    def getflow(self, phase, *species, units='kmol/hr'):
        """Get flow rates of species in given units."""
        index = self.indices(*species) if species else ...
        p = phase_index[phase]
        if units == 'kmol/hr': return self._molarray[p, index]
        flow_wt_units = _Q(1, units)
        dim = flow_wt_units.dimensionality
        if dim == mol_flow_dim:
            return self._molarray[p, index]*flow_wt_units.to(self.units['mol']).magnitude
        elif dim == mass_flow_dim:
            return self._massarray[p, index]*flow_wt_units.to(self.units['mass']).magnitude
        elif dim == vol_flow_dim:
            return self._volarray[p, index]*flow_wt_units.to(self.units['vol']).magnitude
        else:
            raise DimensionError(f"Dimensions for flow units must be in molar, mass or volumetric flow rates, not '{dim}'.")
        
    def setflow(self, phase, flow=(), species=(), units='kmol/hr',
                 inplace='', **flow_pairs):
        """Set flow rates according to the species order and flow_pairs. `inplace` can be any operation that can be performed in place (e.g. +, -, *, /, |, **, etc.)."""
        if species and isinstance(species[0], str):
            species = [self._species_dict[ID] for ID in species]
        species = (*species, *flow_pairs.keys())
        flow = (*flow, *flow_pairs.values())
        p = phase_index[phase]
        index = self.indices(*species) if species else ... 
        if units == 'kmol/hr':
            exec(f"self._molarray[p, index] {inplace}= flow", locals())
            return
        flow_wt_units = _Q(flow, units)
        dim = flow_wt_units.dimensionality
        if dim == mol_flow_dim:
            exec(f"self._molarray[p, index] {inplace}= flow_wt_units.to(self.units['mol']).magnitude", locals())
        elif dim == mass_flow_dim:
            exec(f"self._massarray[p, index] {inplace}= flow_wt_units.to(mass_units).magnitude", locals())
        elif dim == vol_flow_dim:
            exec(f"self._volarray[p, index] {inplace}= flow_wt_units.to(vol_units).magnitude", locals())
        else:
            raise DimensionError(f"Dimensions for flow units must be in molar, mass or volumetric flow rates, not '{dim}'.")

    def _setflows(self, molarray):
        """ Set molar, mass and volumetric flow rates for all phases.
        
        **Parameters**
        
            **mol_array:** Array of molar flow rates (kmol/hr) with species by column and phases ('s', 'l', 'L', 'g') by row.
        
        """
        MW = self._MW
        num_compounds = self._num_compounds
        
        # Molar flow rates
        self._molarray = molarray
        self._phases_molflow = dict(zip(phases, molarray))
        molarray = molarray.view(material_array)
        for mol, ID in zip(molarray, molflows_ID):
            setattr(self, ID, mol)
        
        massflows = [] # Mass flow rates    
        volflows = [] # Volumetric flow rates    
        for phase, mol in zip(phases, molarray):
            vol = []
            mass = []
            for i, s in num_compounds:
                mol_i = mol[i:i+1]
                m = MassFlow(s.ID, (mol_i, MW[i]))
                v = VolumetricFlow(s, (self, mol_i, phase.lower()))
                mass.append(m)
                vol.append(v)
            massflows.append(mass)
            volflows.append(vol)
        
        self._massarray = property_array(massflows)
        self._volarray = property_array(volflows)
        for ID_mass, ID_vol, mass, vol in zip(massflows_ID, volflows_ID, self._massarray, self._volarray):
            setattr(self, ID_mass, mass)
            setattr(self, ID_vol, vol)

    @property
    def phase(self):
        """String denoting all phases present"""
        phase = ''
        if (self._solid_mol != 0).any(): phase += 's'
        if (self._liquid_mol != 0).any(): phase += 'l'
        if (self._LIQUID_mol != 0).any(): phase += 'L'
        if (self._vapor_mol != 0).any(): phase += 'g'
        if phase == '': phase = self._default_phase
        return phase

    @phase.setter
    def phase(self, phase):
        if phase not in phases:
            raise ValueError("Phase must be one of the following: 's', 'l', 'L' or 'g'")
        mol = self._molarray.sum(0)
        self._default_phase = phase
        self.empty()
        self._phases_molflow[phase][:] = mol

    ### Solids ###

    # Molar
    @property
    def solid_mol(self):
        """solid phase molar flow rates (kmol/hr)."""
        return self._solid_mol

    @solid_mol.setter
    def solid_mol(self, val):
        self._solid_mol[:] = val

    @property
    def solid_molfrac(self):
        """solid phase molar fractions."""
        return fraction(self._solid_mol).view(tuple_array)

    @property
    def solid_molnet(self):
        """solid phase net molar flow rate (kg/hr)."""
        return self._solid_mol.sum()

    # Mass
    @property
    def solid_mass(self):
        """solid phase mass flow rates (kg/hr)."""
        return self._solid_mass

    @solid_mass.setter
    def solid_mass(self, val):
        self._solid_mass[:] = val

    @property
    def solid_massfrac(self):
        """solid phase mass fractions."""
        return fraction(self._solid_mol * self._MW).view(tuple_array)

    @property
    def solid_massnet(self):
        """solid phase net mass flow rate (kg/hr)."""
        return self._solid_mass.sum()

    # Volumetric
    @property
    def solid_vol(self):
        """solid phase volumetric flow rates (m3/hr)."""
        return self._solid_vol

    @solid_vol.setter
    def solid_vol(self, val):
        self._solid_vol[:] = val

    @property
    def solid_volfrac(self):
        """solid phase volumetric fractions."""
        return fraction(self._solid_vol).view(tuple_array)

    @property
    def solid_volnet(self):
        """solid phase net volumetric flow rate (m3/hr)."""
        return self._solid_vol.sum()

    ### Liquids ###

    # Molar
    @property
    def liquid_mol(self):
        """liquid phase molar flow rates (kmol/hr)."""
        return self._liquid_mol

    @liquid_mol.setter
    def liquid_mol(self, val):
        self._liquid_mol[:] = val

    @property
    def liquid_molfrac(self):
        """liquid phase molar fractions."""
        return fraction(self._liquid_mol).view(tuple_array)

    @property
    def liquid_molnet(self):
        """liquid phase net molar flow rate (kg/hr)."""
        return self._liquid_mol.sum()

    # Mass
    @property
    def liquid_mass(self):
        """liquid phase mass flow rates (kg/hr)."""
        return self._liquid_mass

    @liquid_mass.setter
    def liquid_mass(self, val):
        self._liquid_mass[:] = val

    @property
    def liquid_massfrac(self):
        """liquid phase mass fractions."""
        return fraction(self._liquid_mass).view(tuple_array)

    @property
    def liquid_massnet(self):
        """liquid phase net mass flow rate (kg/hr)."""
        return self._liquid_mass.sum()

    # Volumetric
    @property
    def liquid_vol(self):
        """liquid phase volumetric flow rates (m3/hr)."""
        return self._liquid_vol

    @liquid_vol.setter
    def liquid_vol(self, val):
        self._liquid_vol[:] = val

    @property
    def liquid_volfrac(self):
        """liquid phase volumetric fractions."""
        return fraction(self._liquid_vol).view(tuple_array)

    @property
    def liquid_volnet(self):
        """liquid phase net volumetric flow rate (m3/hr)."""
        return self._liquid_vol.sum()

    ### LIQUID ###

    # Molar
    @property
    def LIQUID_mol(self):
        """LIQUID phase molar flow rates (kmol/hr)."""
        return self._LIQUID_mol

    @LIQUID_mol.setter
    def LIQUID_mol(self, val):
        self._LIQUID_mol[:] = val

    @property
    def LIQUID_molfrac(self):
        """LIQUID phase molar fractions."""
        return fraction(self._LIQUID_mol).view(tuple_array)

    @property
    def LIQUID_molnet(self):
        """LIQUID phase net molar flow rate (kg/hr)."""
        return self._LIQUID_mol.sum()

    # Mass
    @property
    def LIQUID_mass(self):
        """LIQUID phase mass flow rates (kg/hr)."""
        return self._LIQUID_mass

    @LIQUID_mass.setter
    def LIQUID_mass(self, val):
        self._LIQUID_mass[:] = val

    @property
    def LIQUID_massfrac(self):
        """LIQUID phase mass fractions."""
        return fraction(self._LIQUID_mol * self._MW).view(tuple_array)

    @property
    def LIQUID_massnet(self):
        """LIQUID phase net mass flow rate (kg/hr)."""
        return self._LIQUID_mass.sum()

    # Volumetric
    @property
    def LIQUID_vol(self):
        """LIQUID phase volumetric flow rates (m3/hr)."""
        return self._LIQUID_vol

    @LIQUID_vol.setter
    def LIQUID_vol(self, val):
        self._LIQUID_vol[:] = val

    @property
    def LIQUID_volfrac(self):
        """LIQUID phase volumetric fractions."""
        return fraction(self._LIQUID_vol).view(tuple_array)

    @property
    def LIQUID_volnet(self):
        """LIQUID phase net volumetric flow rate (m3/hr)."""
        return self._LIQUID_vol.sum()

    ### Vapor ###

    # Molar
    @property
    def vapor_mol(self):
        """vapor phase molar flow rates (kmol/hr)."""
        return self._vapor_mol

    @vapor_mol.setter
    def vapor_mol(self, val):
        self._vapor_mol[:] = val

    @property
    def vapor_molfrac(self):
        """vapor phase molar fractions."""
        return fraction(self._vapor_mol).view(tuple_array)

    @property
    def vapor_molnet(self):
        """vapor phase net molar flow rate (kg/hr)."""
        return self._vapor_mol.sum()

    # Mass
    @property
    def vapor_mass(self):
        """vapor phase mass flow rates (kg/hr)."""
        return self._vapor_mass

    @vapor_mass.setter
    def vapor_mass(self, mass):
        self._vapor_mass[:] = mass

    @property
    def vapor_massfrac(self):
        """vapor phase mass fractions."""
        return fraction(self._vapor_mol * self._MW).view(tuple_array)

    @property
    def vapor_massnet(self):
        """vapor phase net mass flow rate (kg/hr)."""
        return self._vapor_mass.sum()

    # Volumetric
    @property
    def vapor_vol(self):
        """vapor phase volumetric flow rates (m3/hr)."""
        return self._vapor_vol

    @vapor_vol.setter
    def vapor_vol(self, vol):
        self._vapor_vol[:] = vol

    @property
    def vapor_volfrac(self):
        """vapor phase volumetric fractions."""
        return fraction(self._vapor_vol).view(tuple_array)

    @property
    def vapor_volnet(self):
        """vapor phase net volumetric flow rate (m3/hr)."""
        return self._vapor_vol.sum()
    
    ### Overall ###

    # Flow rates
    @property
    def mol(self):
        """Molar flow rates (kmol/hr)."""
        return self._molarray.sum(0).view(tuple_array)

    @property
    def mass(self):
        """Mass flow rates (kmol/hr)."""
        return (self._molarray.sum(0) * self._MW).view(tuple_array)
    
    @property
    def vol(self):
        """Volumetric flow rates (kmol/hr)."""
        return self._volarray.sum(0).view(tuple_array)

    # Fractions
    @property
    def molfrac(self):
        """Molar fractions."""
        return fraction(self._molarray.sum(0)).view(tuple_array)

    @property
    def massfrac(self):
        """Mass fractions."""
        return fraction(self._molarray.sum(0) * self._MW).view(tuple_array)

    @property
    def volfrac(self):
        """Volumetric fractions."""
        return fraction(self._volarray.sum(0)).view(tuple_array)

    # Net flow rates
    @property
    def molnet(self):
        """Mol net flow rate (kmol/hr)."""
        return self._molarray.sum()

    @property
    def massnet(self):
        """Mass net flow rate (kmol/hr)."""
        return (self._molarray.sum(0)*self._MW).sum()

    @property
    def volnet(self):
        """Volumetric net flow rate (kmol/hr)."""
        return self.solid_volnet + self.liquid_volnet + self.LIQUID_volnet + self.vapor_volnet

    # Other properties
    @property
    def mu(self):
        # Katti, P.K.; Chaudhri, M.M. (1964). "Viscosities of Binary Mixtures of Benzyl Acetate with Dioxane, Aniline, and m-Cresol". Journal of Chemical and Engineering Data. 9 (1964): 442–443.
        molnet = self.molnet
        if self.molnet == 0:
            return 0
        N = self._Nspecies
        mol = self._molarray.flatten()
        mus = np.zeros(N*4)
        Vms = np.zeros(N*4)
        start = 0
        end = N
        for phase in phases:
            mus[start:end] = self._phaseprop_list('mu', phase)
            Vms[start:end] = self._phaseprop_list('Vm', phase)
            start += N
            end += N
        pos = np.where(mol != 0)
        mol = mol[pos]
        molfrac = mol/molnet
        return exp(sum(molfrac*ln(mus[pos]*Vms[pos])))/self.Vm
    mu.__doc__ = Stream.mu.__doc__
    
    ### Material Properties ###

    # General for a given phase
    def _phaseprop_list(self, prop_ID, phase):
        """Return component property lists for given phase."""
        getattr_ = getattr
        out = np.zeros(self._Nspecies)
        phase = phase.lower()
        P = self.P
        T = self.T
        ic = self._num_compounds.__iter__().__next__
        for m in self._phases_molflow[phase]:
            if m:
                i, c = ic()
                c.P = P
                c.T = T
                c.phase = phase
                out[i] = getattr_(c, prop_ID)
            else: ic()
        return out

    def _phaseprop_molar_flow(self, prop_ID, phase):
        """Return array of component properties * kmol/hr for given phase."""
        return self._phaseprop_list(prop_ID, phase) * self._phases_molflow[phase]

    def _phaseprop_molar_flownet(self, prop_ID, phase):
        """Return sum of component properties * kmol/hr for given phase."""
        return self._phaseprop_molar_flow(prop_ID, phase).sum()

    def _phaseprop_molar_mean(self, prop_ID, phase):
        """Return molar weighted average property for given phase."""
        ppfn = self._phaseprop_list(prop_ID, phase) * self._phases_molflow[phase]
        return ppfn / ppfn.sum()

    # General for all phases
    def _prop_molar_flow(self, prop_ID):
        """Return array of component properties * kmol/hr."""
        ppf = self._phaseprop_molar_flow
        return ppf(prop_ID, 's') + ppf(prop_ID, 'l') + ppf(prop_ID, 'L') + ppf(prop_ID, 'g')

    def _prop_molar_flownet(self, prop_ID):
        """Return sum of component properties * kmol/hr"""
        return sum(self._prop_molar_flow(prop_ID))

    def _prop_molar_mean(self, prop_ID):
        """Return molar weighted average property"""
        return self._prop_molar_flownet(prop_ID) / self.molnet

    ### Specifications ###

    def copylike(self, stream):
        """Copy mol, T, P, and phase of stream to self."""
        if self._species is not stream._species:
            raise ValueError('species must be the same to copy stream specifications.')
        if isinstance(stream, MixedStream):
            self._molarray[:] = stream._molarray
        elif isinstance(stream, Stream):
            self.empty()
            self._phases_molflow[stream.phase][:] = stream.mol
        else:
            raise TypeError('Must pass a valid Stream instance')
        self.P = stream.P
        self.T = stream.T

    ### Equilibrium ###

    def disable_phases(self, phase):
        """Cast stream into a Stream object.
        
        **Parameters**
        
            **phase:** {'s', 'l', 'g'} desired phase of stream
            
        """
        mol = self.mol
        self.__class__ = Stream
        self.__init__(ID=self.ID, flow=mol, T=self.T, P=self.P, phase=phase)

    def enable_phases(self):
        """Cast stream into a MixedStream object."""

    # Vapor-liquid equilibrium
    def VLE(self, species_IDs=None, LNK=None, HNK=None, P=None,
            Qin=None, T=None, V=None, x=None, y=None):
        """Partition flow rates into vapor and liquid phases by equilibrium. Pressure defaults to current pressure.

        **Optional parameters**

            **P:** [float] Operating pressure (Pa)
                
            **species_IDs:** [tuple] IDs of equilibrium species
                 
            **LNK:** tuple[str] Light non-keys
    
            **HNK:** tuple[str] Heavy non-keys

            **User may define one:**
                * **Qin:** [float] Energy input (kJ/hr)
                * **T:** [float] Operating temperature (K)
                * **V:** [float] Overall molar vapor fraction
                * **x:** [numpy array] Molar composition of liquid (for binary mixture)
                * **y:** [numpy array] Molar composition of vapor (for binary mixture)     

        .. Note:
           LNK and HNK are not taken into account for equilibrium. Assumes constant pressure if no pressure is provided.

        """
        ### Decide what kind of equilibrium to run ###
        
        Nspecs = 0
        for i in (P, T, x, y, V, Qin):
            Nspecs += (i is not None)
            
        raise_error = False
        if Nspecs == 0:
            Qin = 0
        elif Nspecs == 2 and not P:
            raise_error = True
        elif Nspecs > 2:
            raise_error = True
        
        if T is not None:
            self.T = T
            eq = 'TP'
        elif x is not None:
            x = np.asarray(x)
            eq = 'xP'
        elif y is not None:
            y = np.asarray(y)
            eq = 'yP'
        elif V is not None:
            eq = 'VP'
        elif Qin is not None:
            eq = 'QinP'
            Hin = self.H + Qin
        else:
            raise_error = True
        
        if raise_error:
            raise ValueError("Invalid specification. Must pass one and only one of the following specifications: T, x, y, V, or Qin.")
        
        if P:
            self.P = P
        else:
            P = self.P
            
        ### Set Up Arguments ###
        
        # Reused attributes
        sp_dict = self._species_dict 
        sp_index = self._IDs.index

        # Set up indices for both equilibrium and non-equilibrium species
        if species_IDs is None:
            species, index = self._equilibrium_species()
            species = tuple(species)
        else:
            index = [sp_index(specie) for specie in species_IDs]
            species = tuple(sp_dict[ID] for ID in species_IDs)
        if LNK:
            LNK_index = [sp_index(specie) for specie in LNK]
        else:
            LNK, LNK_index = self._light_species()
        if HNK:
            HNK_index = [sp_index(specie) for specie in HNK]
        else:
            HNK, HNK_index = self._heavy_species()

        # Get flow rates
        liquid_mol = self._liquid_mol
        vapor_mol = self._vapor_mol
        all_mol = liquid_mol + vapor_mol
        mol = all_mol[index]

        # Set light and heavy keys
        vapor_mol[LNK_index] = all_mol[LNK_index]
        vapor_mol[HNK_index] = 0
        liquid_mol[HNK_index] = all_mol[HNK_index]
        liquid_mol[LNK_index] = 0

        ### Single component equilibrium case ###
        
        Nspecies = len(species)
        if Nspecies == 1:
            # Equilibrium based on one specie
            s = species[0]
            
            if eq == 'TP':
                # Either liquid or gas
                Psat = s.VaporPressure(T)
                if P < Psat:
                    liquid_mol[index] = 0
                    vapor_mol[index] = mol
                else:
                    liquid_mol[index] = mol
                    vapor_mol[index] = 0
                return
            
            elif eq == 'VP':
                # Set vapor fraction
                self.T = s.Tsat(P)
                vapor_mol[index] = mol*V
                liquid_mol[index] = mol - vapor_mol[index]
                return
            
            elif eq == 'QinP': 
                # Set temperature in equilibrium
                self.T = s.Tsat(P)
                
                # Check if super heated vapor
                vapor_mol[index] = mol
                liquid_mol[index] = 0
                H_dew = self.H
                if Hin >= H_dew:
                    self.H = Hin
                    return
    
                # Check if subcooled liquid
                vapor_mol[index] = 0
                liquid_mol[index] = mol
                H_bubble = self.H
                if Hin <= H_bubble:
                    self.H = Hin
                    return
                
                # Adjust vapor fraction accordingly
                V = (Hin - H_bubble)/(H_dew - H_bubble)
                vapor_mol[index] = mol*V
                liquid_mol[index] = mol - vapor_mol[index]
                return
                
            else:
                ValueError("Invalid specification. Pure component equilibrium requires one and only one of the following specifications: T, V, or Qin.")
                
        ### Begin multi-component equilibrium ###
        
        # Get overall composition
        molnet = mol.sum()
        if molnet == 0:
            return  # No equilibrium necessary
        else:
            zf = mol/molnet
        
        min_ = np.zeros(Nspecies)

        ### Get Equilibrium ###

        if eq == 'xP' or eq == 'yP':
            # Get temperature based on phase composition
            if eq == 'xP':
                self.T, y = self._bubble_point(species, x, P)
            else:
                self.T, x = self._dew_point(species, y, P)
                
            if Nspecies > 2:
                raise ValueError(f'More than two components in equilibrium. Only binary component equilibrium can be solved for specification, {eq}.')
            # Get flow rates based on lever rule
            split_frac = (zf[0]-x[0])/(y[0]-x[0])
            if split_frac > 1.0001 or split_frac < -0.0001:
                raise EquilibriumError('Desired composition is not feasible')
            elif split_frac > 1:
                split_frac = 1
            elif split_frac < 0:
                split_frac = 0
            v_net = molnet * split_frac
            vapor_mol[index] = v_net*np.array([y[0], 1 - y[0]])
            liquid_mol[index] = mol - vapor_mol[index]
            return

        # Need to find these for remainding equilibrium types
        T_dew, x_dew = self._dew_point(species, zf, P)
        T_bubble, y_bubble = self._bubble_point(species, zf, P)

        # Guestimate vapor composition for 'VP', 'TP', 'QinP' equilibrium
        y = self._y_cached.get(species)

        def v_given_TPmol(v_guess, T, P, mol):
            """Return equilibrium vapor flow rates given T, P and molar flow rates."""
            Psat_P = np.array([s.VaporPressure(T) for s in species])/P
            actcoef = self.activity_coefficients
            def v_error(v):
                """Error function for constant T and P, where v represents vapor flow rates."""
                l = mol - v
                x = l/l.sum()
                y = v/v.sum()
                err = x*Psat_P * actcoef(species, x, T) - y
                return abs(err)
            
            return least_squares(v_error, v_guess,
                                 bounds=(min_, mol),
                                 ftol=0.001).x

        if eq == 'VP':
            def V_error(T):
                nonlocal v
                if T >= T_dew:
                    return V - 1 - (T - T_dew)
                elif T <= T_bubble:
                    return V + (T - T_bubble)
                v = v_given_TPmol(v, T, P, mol)
                return (V - sum(v)/molnet)
            
            # Guess vapor flow rates
            pos = np.where(mol <= 0)
            mol[pos] = 10**-9
            if y is None: y = V*zf + (1-V)*y_bubble
            v = y*V*molnet
            
            # Solve
            if not hasattr(self, '_T_VP'):
                self._T_VP = (1-V)*T_bubble + V*T_dew
            try:
                self._T_VP = self.T = newton(V_error, self._T_VP)
            except:
                self._T_VP = self.T = brentq(V_error, T_bubble, T_dew)
            
            v[pos] = 0
            vapor_mol[index] = v
            liquid_mol[index] = mol - vapor_mol[index]
            return
        
        elif eq == 'TP':
            if y is None:
                # Guess composition in the vapor is a weighted average of bubble/dew points
                f = (T - T_dew)/(T_bubble - T_dew)
                y = f*zf + (1-f)*y_bubble
            
            # Check if there is equilibrium
            if T > T_dew:
                vapor_mol[index] = v = mol
            elif T < T_bubble:
                vapor_mol[index] = v = min_
            else:
                # Guess overall vapor fraction
                V_guess = (T - T_bubble)/(T_dew - T_bubble)

                # Guess vapor flow rates
                v_guess = y * V_guess * mol

                # Solve
                vapor_mol[index] = v = v_given_TPmol(v_guess, T, P, mol)
                self._y = v/sum(v)
            liquid_mol[index] = mol - v
            return
        
        elif eq == 'QinP':
            # Check if super heated vapor
            vapor_mol[index] = mol
            liquid_mol[index] = 0
            self.T = T_dew
            H_dew = self.H
            if Hin >= 1.01*H_dew:
                self.H = Hin
                return

            # Check if subcooled liquid
            vapor_mol[index] = 0
            liquid_mol[index] = mol
            self.T = T_bubble
            H_bubble = self.H
            if Hin <= 0.99*H_bubble:
                self.H = Hin
                return

            def H_error_T(T):
                nonlocal v
                self.T = T
                vapor_mol[index] = v = v_given_TPmol(v, T, P, mol)
                liquid_mol[index] = mol - v
                return abs(Hin - (self.H))

            # Guess T, overall vapor fraction, and vapor flow rates
            if not hasattr(self, '_T_QinP'):
                self._T_QinP = T_bubble + (Hin - H_bubble)/(H_dew - H_bubble) * (T_dew - T_bubble)
            V_guess = (self._T_QinP - T_bubble)/(T_dew - T_bubble)
            if y is None:
                # Guess composition in the vapor is a weighted average of boiling points
                f = (self._T_QinP - T_dew)/(T_bubble - T_dew)
                y = f*zf + (1-f)*y_bubble
            v = y * V_guess * mol
            
            # Solve
            try:
                try:
                    self._T_QinP = self.T = newton(H_error_T, self._T_QinP)
                except:
                    self._T_QinP = T_bubble + (Hin - H_bubble)/(H_dew - H_bubble) * (T_dew - T_bubble)
                    V_guess = (self._T_QinP - T_bubble)/(T_dew - T_bubble)
                    f = (self._T_QinP - T_dew)/(T_bubble - T_dew)
                    y = f*zf + (1-f)*y_bubble
                    v = y * V_guess * mol
                    self._T_QinP = self.T = newton(H_error_T, self._T_QinP)
            except:
                self._T_QinP = self.T = minimize_scalar(H_error_T,
                                                        bounds=(T_bubble, T_dew),
                                                        method='bounded').x
                
            self._y_cached[species] = v/v.sum()

    # LIQUID-liquid equilibrium
    def LLE(self, species_IDs=None, split=None, lNK=(), LNK=(),
            solvents=(), solvent_split=(),
            P=None, T=None, Qin=0):
        """Partition flow rates into liquid and LIQUID phases by equilibrium.

        **Optional Parameters**

            **species_IDs:** *tuple[str]* IDs of equilibrium species
            
            **split:** *tuple[float]* Initial guess split fractions of each specie to the 'liquid'

            **lNK:** *tuple[str]* Species assumed to completely remain in the 'liquid' phase.

            **LNK:** *tuple[str]* Species assumed to completely remain in the 'LIQUID' phase.

            **solvents:** *tuple[str]* Species corresponding to specified solvent_split

            **solvent_split:** *tuple[float]* Split fractions of each specie to the 'liquid' phase.                
            
            **Qin:** *[float]* Energy input (kJ/hr)
            
            **T:** *[float]* Operating temperature (K)
            
            **P:** *[float]* Operating pressure (Pa)    

        .. Note:
           lNK and LNK are not taken into account for equilibrium. Assumes constant pressure and temperatrue if none are provided.

        """
        ### Set Up ###

        # Set up indices for both equilibrium and non-equilibrium species
        sp_dict = self._species_dict
        sp_index = self._IDs.index
        if species_IDs is None:
            species, index = self._equilibrium_species()
        else:
            species = [sp_dict[ID] for ID in species_IDs]
            index = [sp_index(ID) for ID in species_IDs]
        
        for s, i in zip(species, index):
            ID = s.ID
            if (ID in solvents) or (ID in lNK) or (ID in LNK):
                species.remove(s)
                index.remove(i)

        species = tuple(species)
        lNK_index = [sp_index(s) for s in lNK]
        LNK_index = [sp_index(s) for s in LNK]
        solvent_index = [sp_index(s) for s in solvents]

        # Get min and max splits
        Nspecies = len(species)
        min_ = np.zeros(Nspecies)

        # Set up activity coefficients
        solvent_species = tuple(sp_dict[ID] for ID in solvents)
        activity_species = species + solvent_species
        act_coef = self.activity_coefficients

        # Make sure splits are given as arrays
        if split is None:
            split = self._lL_split_cached.get(species)
            if split is None:
                split_dipoles = [s.dipole for s in species]
                solvent_dipoles = [s.dipole for s in solvent_species]
                dipoles = split_dipoles + solvent_dipoles
                split_dipoles = np.array(split_dipoles)
                try:
                    split = split_dipoles/max(dipoles)
                except TypeError as TE:
                    missing_dipoles = []
                    for i, is_missing in enumerate(split_dipoles==None):
                        if is_missing:
                            missing_dipoles.append(activity_species[i])
                    raise EquilibriumError(f'Cannot make estimate for split. Missing dipole values for species: {missing_dipoles}')
                split = np.array(split)
            solvent_split = np.array(solvent_split)

        # Equilibrium by constant temperature or with duty
        if T:
            self.T = T
        elif Qin:
            self.H += Qin
            T = self.T
        else:
            T = self.T
        
        if P:
            self.P = P
        else:
            P = self.P

        # Get molar flow rates
        liquid_mol = self._liquid_mol
        LIQUID_mol = self._LIQUID_mol
        all_mol = liquid_mol + LIQUID_mol

        # Equilibrium species
        mol = all_mol[index]  # Solute species flow rates
        Kmol = all_mol[solvent_index]  # Solvent species flow rates
        Kguess = Kmol * solvent_split  # Solvent species flow rates to 'l' phase
        molKmol = np.concatenate((mol, Kmol))  # Both

        # Non-equilibrium species
        lNK_mol = all_mol[lNK_index]
        LNK_mol = all_mol[LNK_index]

        ### Solve ###
        
        # Error function
        def guess_error(l_guess):
            # Error function for constant T and P, where l represents 'l' liquid flow rates
            l_guess = np.concatenate((l_guess, Kguess))
            L_guess = molKmol - l_guess
            x1_guess = l_guess/sum(l_guess)
            x2_guess = L_guess/sum(L_guess)
            x1_gamma = act_coef(activity_species, x1_guess, T)
            x2_gamma = act_coef(activity_species, x2_guess, T)
            err = (x1_guess * x1_gamma - x2_guess * x2_gamma)
            return abs(err)[0:Nspecies]

        # Guestimates
        pos = mol == 0
        mol[pos] = 10**-20  # Give it a little to not hit bounds for least squares
        l_guess = split * mol  # Initial guess

        # Solve
        sol = least_squares(guess_error, l_guess, bounds=(min_, mol))
        l_guess = sol.x
        l_guess[pos] = 0
        split = l_guess/mol
        
        # Make sure two phases are given
        xl = l_guess/sum(l_guess)
        L_guess = mol-l_guess
        xL = L_guess/sum(L_guess)
        if (xl - xL < 0.1).all():
            raise EquilibriumError('Could not solve equilibrium, please input better split guesses')

        # Set results
        self._lL_split_cached[species] = split
        liquid_mol[index] = l_guess
        liquid_mol[lNK_index] = lNK_mol
        liquid_mol[LNK_index] = 0
        liquid_mol[solvent_index] = Kmol * solvent_split
        LIQUID_mol[index] = mol-l_guess
        LIQUID_mol[lNK_index] = 0
        LIQUID_mol[LNK_index] = LNK_mol
        LIQUID_mol[solvent_index] = Kmol - liquid_mol[solvent_index]

    def _info(self, **show_units):
        """Return string with all specifications."""
        units = self.units
        basic_info = self._info_header() + '\n'
        T_units, P_units, flow_units, fraction = self._info_units(show_units)
        phases = self.phase
        basic_info += self._info_phaseTP(phases, T_units, P_units)
        
        # Get flow attribute name
        flow_dim = _Q(0, flow_units).dimensionality
        if flow_dim == mol_flow_dim:
            flow_type = 'mol'
        elif flow_dim == mass_flow_dim:
            flow_type = 'mass'
        elif flow_dim == vol_flow_dim:
            flow_type = 'vol'
        else:
            raise DimensionError(f"Dimensions for flow units must be in molar, mass or volumetric flow rates, not '{flow_dim}'.")
        
        def flow_rate_skeleton(phase, flow_type, flow_units, fraction, maxlen):
            phase_full_name = letter_phases[phase]
            beginning = f' {phase_full_name}: '
            new_line_spaces = len(beginning) * ' '

            if fraction:
                flowname = flow_type + 'net'
                flownet = getattr(self, phase_full_name + '_' + flowname)
                end = ('\n' + new_line_spaces + 'net' +(maxlen-3)*' ' + '  ' +
                       f'{flownet:.3g} {flow_units}')
            else:
                end = ''

            return beginning, new_line_spaces, end

        # Length of species column
        nonzero, species = self.nonzero_species
        if len(nonzero) == 0:
            return basic_info + ' flow: 0'
        all_lengths = [len(sp) for sp in species]
        maxlen = max(all_lengths + [7])  # include length of the word 'species'

        # Set up flow rate information for all phases
        first_phase = True
        phases_flowrates_info = ''
        for phase in phases:
            # Get flow rates of phase in nice structure
            flow_attr = letter_phases[phase] + '_' + flow_type
            if fraction: flow_attr += 'frac'
            flow = getattr(self, flow_attr)
            nonzero, species = nonzero_species(self._num_IDs, flow)
            if fraction:
                flow_nonzero = flow[nonzero]
            else:
                flow_nonzero = _Q(flow[nonzero], units[flow_type]).to(flow_units).magnitude
            
            # Get basic structure for phase flow rates
            beginning, new_line_spaces, end = flow_rate_skeleton(phase, flow_type, flow_units, fraction, maxlen)

            # Make up for soild and vapor phase having 5 letters
            # (both liquid and LIQUID have 6 letters)
            if phase in 'sg':
                beginning += ' '
                new_line_spaces += ' '
                if fraction: end = '\n ' + end[1:]

            # Set flow rates info
            flowrates = ''
            l = len(nonzero)
            lengths = [len(sp) for sp in species]
            for i in range(l-1):
                spaces = ' ' * (maxlen - lengths[i])
                flowrates += f'{species[i]} ' + spaces + \
                    f' {flow_nonzero[i]:.3g}\n' + new_line_spaces
            spaces = ' ' * (maxlen - lengths[l-1])
            flowrates += (f'{species[l-1]} ' + spaces
                          + f' {flow_nonzero[l-1]:.3g}')

            # Add header to flow rates
            if first_phase:
                spaces = ' ' * (maxlen - 7)
                beginning = (f'{new_line_spaces}species{spaces}  {flow_units}\n'
                             + beginning)

            # Put it together
            phases_flowrates_info += beginning + flowrates + end + '\n\n'
            first_phase = False

        return basic_info + phases_flowrates_info[:-2]


Stream.VLE.__doc__ = MixedStream.VLE.__doc__
Stream.LLE.__doc__ = MixedStream.LLE.__doc__
MixedStream.mu.__doc__ = Stream.mu.__doc__





# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 13:42:33 2018

@author: yoelr
"""
from . import compounds as comps

Chemical = comps.Chemical
Compound = comps.Compound
__all__ = ('Species',)

# TODO: Fix how Dortmund groups are selected in thermo.Chemical. Glycerol defaults to wrong value.

# %% Managing ChEDL thermodynamic properties

class Species:
    """Create Species object that contains Compound objects as attributes.

    **Parameters**

        ***IDs:** [str] Strings should be one of the following [-]:
                  * Name, in IUPAC form or common form or a synonym registered in PubChem
                  * InChI name, prefixed by 'InChI=1S/' or 'InChI=1/'
                  * InChI key, prefixed by 'InChIKey='
                  * PubChem CID, prefixed by 'PubChem='
                  * SMILES (prefix with 'SMILES=' to ensure smiles parsing)
                  * CAS number
         
        **Optional**
        
        **cls:** [type Compound] The type of objects that will be created and stored as attributes. Defaults to biosteam.Chemical
        
    """
    units = Compound.units
    _immutable = set()
    
    @classmethod
    def tospecies(cls, compounds, copy=True):
        """Return a Species object from an iterable of Compound objects."""
        if not copy and isinstance(compounds, Species): return compounds
        self = cls.__new__(cls)
        setter = super(cls, self).__setattr__
        for c in compounds: 
            if isinstance(c, Compound): setter(c.ID, c)
        return self
    
    def __init__(self, *ID, cls=None):
        # Sphinx docs look ugly if a class is in the default argument
        if cls is None: cls = Chemical
        
        # Make sure ID is a tuple
        if not ID:
            return
        elif isinstance(ID, str):
            ID = (ID,)
        elif not isinstance(ID[0], str):
            ID = ID[0] # For backwards compatibility
        
        # Set Compound object attributes
        attrset = super().__setattr__
        for n in ID:
            try:
                compound = cls(n)
            except:
                raise Exception(f"Compound, '{n}', not defined in databank.")
            attrset(n, compound)

    def __setattr__(self, ID, compound):
        if self in self._immutable:
            raise ValueError('Cannot alter species object attached to Stream objects.')
        if isinstance(compound, Compound):
            compound.ID = ID
            super().__setattr__(ID, compound)
        elif ID == 'ID':
            new_IDs = compound
            new_species = [getattr(self, i) for i in new_IDs]
            self.__dict__.clear()
            for ID, compound in zip(new_IDs, new_species):
                super().__setattr__(ID, compound)
        else:
            raise TypeError('Can only set Compound objects as attributes')
    
    def __delattr__(self, ID):
        if self in self._immutable:
            raise ValueError('Cannot alter species object attached to Stream objects.')
        else:
            super().__delattr__(ID)
    
    def __iter__(self):
        return iter(self.__dict__.values())
    
    @property
    def CAS(self):
        return (s.CAS for s in self)
    
    @property
    def ID(self):
        return iter(self.__dict__.keys())

    def getprops(self, species_IDs, prop_ID, T=None, P=None, phase=None) -> list:
        """Return list of the desired property, prop_ID, for each compound in species_ID."""
        if isinstance(species_IDs, str):
            species = (getattr(self, species_IDs),)
        else:
            species = (getattr(self, i) for i in species_IDs)
        return self._getprops(species, prop_ID, T, P, phase)

    @staticmethod
    def _getprops(species, prop_ID, T, P, phase) -> list:
        """Return list of the desired property, prop_ID, for given species."""
        props = []; sat = setattr; gat = getattr
        for s in species:
            if T: sat(s, 'T', T)
            if P: sat(s, 'P', P)
            if phase: sat(s, 'phase', phase)
            props.append(gat(s, prop_ID))
        return props

    def setprops(self, species_IDs, prop_ID, new_values):
        """Set new values to specie property, prop_ID, for given species_IDs."""
        sat = setattr; gat = getattr
        for ID in species_IDs: 
            sat(gat(self, ID), prop_ID, new_values)

    # Representation
    def _info(self):
        """Information on self"""
        return 'Species: \n ' + str(list(self.ID)).replace("'", '')[1:-1]

    def show(self):
        """print information on self"""
        print(self._info())

    def __repr__(self):
        return '<Species: ' + str(list(self.ID)).replace("'", '')[1:-1] + '>'

    



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:14:01 2018

@author: yoelr
"""
from .. import Unit
from . import Splitter


class CrushingMill(Unit):
    _kwargs = Splitter._kwargs
    _run = Splitter._run
    _has_power_utility = True
    
    #: Original Price (USD)
    C_0 = 1.5e6
    
    #: Original flow rate (kg/hr)
    V_0 = 335e3 
    
    #: Scaling exponent
    exp = 0.60 
    
    #: Electricity rate (kW/(kg/hr))
    electricity_rate = 0.006
    
    #: Original CEPCI 
    CEPCI_0 = 541.7
    
    def _cost(self):
        # Size factor
        massflow = self.ins[0].massnet
        S = massflow/self.V_0
        self._results['Cost']['Crushing mill'] = (self.CEPCI/self.CEPCI_0) * self.C_0*S**self.exp
        self._power_utility(massflow*self.electricity_rate)



""" A module of tables and graphs for constants
"""
from math import log, pi

__all__ = ('GTable', 'HNATable', 'FinalValue',
           'VesselWeightAndWallThickness', 'Kvalue')

def GTable(DRho, Hlr):
    """Return the allowable downflow (baffle liquid load) in gph/ft2, usually used for vertical vessel.
    
    **Parameters**
    
        DRho: [float] Density difference between light liquid and vapor (lb/ft^3 ?)
        
        Hlr: [float] Height of liquid level above the interphase of light liquid and heavy liquid (ft)
    
    """

    A = {}
    B = {}
    C = {}
    D = {}

    # TODO: Add errors when outside the range!
    if Hlr > 30.0:
        Hlr = 30.0

    if Hlr < 18.0:
        Hlr = 18.0

    if DRho > 50.0:
        DRho = 50.0

    if DRho < 10.0:
        DRho = 10.0
    
    Hlr = round(Hlr, 0)

    A[18] = -9000.0
    B[18] = 1275.4
    C[18] = -31.3571
    D[18] = 0.255556

    A[19] = -4690.0
    B[19] = 900.117
    C[19] = -20.5252
    D[19] = 0.157254

    A[20] = -9980.0
    B[20] = 1367.91
    C[20] = -33.0163
    D[20] = 0.26476

    A[21] = -8120.0
    B[21] = 1147.46
    C[21] = -25.3
    D[21] = 0.184444

    A[22] = -16800.0
    B[22] = 1964.89
    C[22] = -48.8627
    D[22] = 0.399498

    A[23] = -7900.0
    B[23] = 1255.35
    C[23] = -29.9142
    D[23] = 0.235632

    A[24] = -11200.0
    B[24] = 1561.48
    C[24] = -38.7335
    D[24] = 0.318511

    A[25] = -11100.0
    B[25] = 1554.66
    C[25] = -38.0313
    D[25] = 0.308026

    A[26] = -7410.0
    B[26] = 1274.0
    C[26] = -30.8013
    D[26] = 0.246585

    A[27] = -12700.0
    B[27] = 1709.78
    C[27] = -42.1048
    D[27] = 0.342222

    A[28] = -10200.0
    B[28] = 1507.78
    C[28] = -36.422
    D[28] = 0.291221

    A[29] = -10700.0
    B[29] = 1553.51
    C[29] = -37.5721
    D[29] = 0.300279

    A[30] = -9830.0
    B[30] = 1513.11
    C[30] = -37.1907
    D[30] = 0.30379

    G = A[Hlr] + (B[Hlr] * DRho) + (C[Hlr] * DRho ** 2) + (D[Hlr] * DRho ** 3)

    return round(G, 2)


def HNATable(Type, X):
    """Table for cylindrical height and area conversions.
        
    **Parameters**
    
        Type: [int] 1 if given H/D and find A/At, 2 if given A/At and find H/D
        
        X: [float] H/D or A/At
        
    """
    # Type = 1 is where H/D is known, find A/At, Type = 2 is where A/At is known, find H/D
    if (Type == 1):
        a = -0.0000475593
        b = 3.924091
        c = 0.174875
        d = -6.358805
        e = 5.668973
        f = 4.018448
        g = -4.916411
        h = -1.801705
        i = -0.145348
        Y = (a + c * X + e * X ** 2 + g * X ** 3 + i * X ** 4) / \
            (1.0 + b * X + d * X ** 2 + f * X ** 3 + h * X ** 4)
    else:
        a = 0.00153756
        b = 26.787101
        c = 3.299201
        d = -22.923932
        e = 24.353518
        f = -14.844824
        g = -36.999376
        h = 10.529572
        i = 9.892851
        Y = (a + c * X + e * X ** 2 + g * X ** 3 + i * X ** 4) / \
            (1.0 + b * X + d * X ** 2 + f * X ** 3 + h * X ** 4)

    return Y


def VesselWeightAndWallThickness(Pressure, Diameter, VesselLength, rho_M, Je=0.85):
    """Return vessel weight and wall thickness.
    
    **Parameters**
    
        Pressure: [float] (psia)
        
        Diameter: [float] (ft)
        
        VesselLength: [float] (ft)
        
        rho_M: [float] Density of Material (lb/ft^3)
        
        Je: [float] Joint efficiency (1.0 for X-Rayed joints, 0.85 for thin carbon steel)
    
    """
    S = 15000.0     # Vessel material stress value (assume carbon-steel)
    Ca = 1.0/8.0    # Corrosion Allowance in inches
    
    P1 = abs(Pressure - 14.7) + 30.0
    P2 = abs(Pressure - 14.7)*1.1
    if P1 > P2:
        PT = P1
    else:
        PT = P2

    # Calculate the wall thickness and surface area
    # Shell
    SWT = (PT * Diameter*12.0) / (2.0 * S * Je - 1.2 * PT) + Ca
    SSA = pi * Diameter * VesselLength
    if Diameter < 15.0 and PT > (100 - 14.7):
        # Elliptical Heads
        HWT = (PT * Diameter*12.0) / (2.0 * S * Je - 0.2 * PT) + Ca
        HSA = 1.09 * Diameter ** 2
    elif Diameter > 15.0:
        # Hemispherical Heads
        HWT = (PT * Diameter*12.0) / (4.0 * S * Je - 0.4 * PT) + Ca
        HSA = 1.571 * Diameter ** 2
    else:
        # Dished Heads
        HWT = 0.885 * (PT * Diameter*12.0) / (S * Je - 0.1 * PT) + Ca
        HSA = 0.842 * Diameter ** 2

    # Approximate the vessel wall thickness, whichever is larger
    if SWT > HWT:
        ts = SWT
    else:
        ts = HWT

    # Minimum thickness for vessel rigidity may be larger
    if Diameter < 4:
        ts_min = 1/4
    elif Diameter < 6:
        ts_min = 5/16
    elif Diameter < 8:
        ts_min = 3/8
    elif Diameter < 10:
        ts_min = 7/16
    elif Diameter < 12:
        ts_min = 1/2
    else:
        ts_min = ts
    if ts < ts_min:
        ts = ts_min
        
    VW = rho_M * ts/12 * (SSA + 2.0 * HSA)  # in lb
    VW = round(VW, 2)
    return VW, ts


def LowLiqLevelHeight(Type, P, D):
    """Table to obtain Hlll for two-phase separators.
    
    **Parameters**
    
        Type: [int] 1 for vertical, 2 for horizontal
        
        P: [float] Pressure (psia)
        
        D: [float] Diameter (ft)
    
    """
    if Type == 1:
        Hlll = 0.5
        if P < 300:
            Hlll = 1.25

    elif Type == 2:
        if D <= 4.0:
            Hlll = 9.0/12.0
        elif D > 4.0 and D <= 7.0:
            Hlll = 10.0/12.0
        elif D > 7.0 and D <= 9.0:
            Hlll = 11.0/12.0
        elif D > 9.0 and D <= 11.0:
            Hlll = 1.0
        elif D > 11.0 and D <= 15.0:
            Hlll = 13.0/12.0
        else:
            Hlll = 15.0/12.0

    return Hlll  # in ft


def Kvalue(P):
    """Return K-constant in Stoke's Law using the York-Demister equation.
    
    **Parameters**
    
        P: [float] Pressure (psia)
        
    """
    if P >= 0 and P <= 15.0:
        K = 0.1821+(0.0029*P)+(0.046*log(P))
    elif P > 15.0 and P <= 40.0:
        K = 0.35
    elif P > 40.0 and P <= 5500.0:
        K = 0.43 - 0.023*log(P)
    else:
        raise ValueError(f'Invalid Pressure {P} psia')
    return K


def FinalValue(value):
    """Return value to the next/highest 0.5 units"""
    intval = round(value)
    if value > intval:
        return intval + 0.5
    elif value == intval:
        return value
    else:
        return intval - 0.5



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:45:47 2018

@author: yoelr
"""
import numpy as np
from ._reactor import BatchReactor
from ._hx import HXutility
from .. import Stream
from scipy.integrate import odeint

class Fermentation(BatchReactor):
    """
    Create a Fermentation object which models large-scale batch fermentation for the production of 1st generation ethanol using yeast [1, 2, 3, 4]. Only sucrose and glucose are taken into account. Conversion is based on reaction time, *tau*. Cleaning and unloading time, *tau_0*, fraction of working volume, *V_wf*, and number of reactors, N_reactors, are attributes that can be changed. Cost of a reactor is based on the NREL batch fermentation tank cost assuming volumetric scaling with a 6/10th exponent [3].
    
    **Parameters**
    
        **tau:** Reaction time.
        
        **efficiency:** User enforced efficiency.
    
    **ins**
    
        [:] Inffluent streams
        
    **outs**
    
        [0] Effluent
        
        [1] CO2
    
    **References**
    
        [1] Oliveira, Samuel C., et al. "Discrimination between ethanol inhibition models in a continuous alcoholic fermentation process using flocculating yeast." Applied biochemistry and biotechnology 74.3 (1998): 161-172.
        
        [2] Oliveira, Samuel C., et al. "Continuous ethanol fermentation in a tower reactor with flocculating yeast recycle: scale-up effects on process performance, kinetic parameters and model predictions." Bioprocess Engineering 20.6 (1999): 525-530.
        
        [3] Oliveira, Samuel C., et al. "Mathematical modeling of a continuous alcoholic fermentation process in a two-stage tower reactor cascade with flocculating yeast recycle." Bioprocess and biosystems engineering 38.3 (2015): 469-479.
        
        [4] Oliveira, Samuel C., et al. "Kinetic Modeling of 1‐G Ethanol Fermentations." Fermentation Processes. InTech, 2017.
        
        [5] D. Humbird, R. Davis, L. Tao, C. Kinchin, D. Hsu, and A. Aden National. Renewable Energy Laboratory Golden, Colorado. P. Schoen, J. Lukas, B. Olthof, M. Worley, D. Sexton, and D. Dudgeon. Harris Group Inc. Seattle, Washington and Atlanta, Georgia. Process Design and Economics for Biochemical Conversion of Lignocellulosic Biomass to Ethanol Dilute-Acid Pretreatment and Enzymatic Hydrolysis of Corn Stover. May 2011. Technical Report NREL/TP-5100-47764
        
    .. note::
        
        A compound with CAS 'Yeast' must be present in species.
        
    **Examples**
    
        :doc:`Fermentation Example`
        
    """
    _units = {'Reactor volume': 'm3',
              'Cycle time': 'hr',
              'Loading time': 'hr',
              'Total dead time': 'hr'}
    _kwargs = {'tau': None, # Reaction time
               'efficiency': None}  # Theoretical efficiency
    _N_heat_utilities = 0
    _has_power_utility = True
    
    #: Cleaning and unloading time (hr)
    tau_0 = 12 
    
    #: Number of reactors
    _N_reactors = None
    
    #: Fraction of filled tank to total tank volume
    V_wf = 0.8
    
    #: Base CEPCI 
    CEPCI_0 = 521.9 
    
    #: Base price (USD)    
    C_0 = 844000
    
    #: Base volume (m3)    
    V_0 = 3785
    
    #: Scaling exponent for costing
    exp = 0.6 
    
    #: Base agitator price (USD)
    C_A = 52500 
    
    #: Base agitator power (kW)
    A_p = 22.371
    
    #: tuple of kinetic parameters for the kinetic model. Default constants are fitted for Oliveria's model (mu_m1, mu_m2, Ks1, Ks2, Pm1, Pm2, Xm, Y_PS, a)
    kinetic_constants = (0.31,  # mu_m1
                         1.01,  # mu_m2
                         1.88,  # Ks1
                         2.81,  # Ks2
                         82.8,  # Pm1
                         108.2, # Pm2
                         113.4, # Xm
                         0.45,  # Y_PS
                         0.18)  # a
    @property
    def N_reactors(self):
        """Enforced Number of batch reactors."""
        return self._N_reactors
    
    @N_reactors.setter
    def N_reactors(self, N):
        if N <= 1: raise ValueError(f"Number of reactors must be greater than 1, value {N} is infeasible")
        self._N_reactors = N

    def _init(self):
        self._cooler = hx = HXutility(None)
        self._heat_utilities = hx._heat_utilities
    
    def _setup(self):
        if not self._kwargs['tau']:
            raise ValueError(f"Reaction time must be larger than 0, not '{self._kwargs['tau']}'")

    def _calc_efficiency(self, feed, tau):
        # Get initial concentrations
        y, e, s, w = feed.indices('Yeast',
                                  '64-17-5',
                                  '492-61-5',
                                  '7732-18-5', CAS=True)
        mass = feed.mass
        volnet = feed.volnet
        concentration_in = mass/volnet
        X0, P0, S0 = (concentration_in[i] for i in (y, e, s))
        
        # Integrate to get final concentration
        t = np.linspace(0, tau, 1000)
        C_t = odeint(self.kinetic_model, (X0, P0, S0), t,
                     args=self.kinetic_constants)
        # Cache data
        self._X = C_t[:, 0]
        self._P = C_t[:, 1]
        self._S = S = C_t[:, 2]
        
        # Calculate efficiency
        Sf = S[-1]
        Sf = Sf if Sf > 0 else 0
        Y_PS = self.kinetic_constants[-2]
        eff = (S0 - Sf)/S0 * Y_PS/0.511
        return eff
        
    @staticmethod
    def kinetic_model(z, t, *kinetic_constants) -> '(dXdt, dPdt, dSdt)':
        """Return change of yeast, ethanol, and substrate concentration in kg/m3.
        
        **Parameters**
        
            **z:** Iterable with (X, E, S) [-]:
                * X: Yeast concentration (kg/m3)
                * P: Ethanol concentration (kg/m3)
                * S: Substrate concentration (kg/m3)
            
            **t:** Time point
            
            ***kinetic_constants:**
                * mu_m1: Maximum specific growth rate (1/hr)
                * mu_m2: Maximum specific ethanol production rate (g-product/g-cell-hr)
                * Ks1: Sugar saturation constant for growth (g/L)
                * Ks2: Sugar saturation constant for product (g/L)
                * Pm1: Maximum product concentration at zero growth [mu_m1=0] (g/L)
                * Pm2: Maximum product concentration [mu_m2=0] (g/L)
                * Xm: Maximum cell concentration [mu_m1=0] (g/L)
                * Y_PS: Ethanol yield based on sugar consumed
                * a: Toxic power
                
        """
        mu_m1, mu_m2, Ks1, Ks2, Pm1, Pm2, Xm, Y_PS, a = kinetic_constants
        
        # Current yeast, ethanol, and glucose concentration (kg/m3)
        X, P, S = z
        
        # Compute coefficients
        mu_X = mu_m1 * (S/(Ks1 + S)) * (1 - P/Pm1)**a*((1-X/Xm))
        mu_P = mu_m2 * (S/(Ks2 + S)) * (1 - P/Pm2)
        mu_S = mu_P/0.45
        
        # Compute derivatives
        dXdt = mu_X * X
        dPdt = (mu_P * X)
        dSdt =  - mu_S * X
        return (dXdt, dPdt, dSdt)
        

    def _run(self):
        # Assume no Glycerol produced
        # No ethanol lost with CO2
        mw_Ethanol = 46.06844
        mw_glucose = 180.156

        # Unpack
        kwargs = self._kwargs
        out, CO2 = self._outs
        out.sum(out, self._ins)
        
        # Species involved in reaction
        species = ['64-17-5', '492-61-5', '57-50-1', '7732-18-5']
        e, g, s, w = indx = out.indices(*species, CAS=True)
        glucose = out.mol[indx[1]]
        sucrose = out.mol[indx[2]]
        
        # Hydrolysis
        new_glucose = sucrose*2
        ch_sucrose = -sucrose
        ch_Water = -sucrose
        out._mol[[g, s, w]] += [new_glucose, ch_sucrose, ch_Water]
        out.T = 32 + 273.15
        
        # Fermentation
        self._results['Design']['Efficiency'] = eff = (
            kwargs.get('efficiency')
            or self._calc_efficiency(out, kwargs['tau'])
            )
        fermented = eff*(glucose + new_glucose)
        ch_glucose = - fermented
        ch_Ethanol = 0.5111*fermented * mw_glucose/mw_Ethanol
        co2 = out.indices('124-38-9', CAS=True)
        changes = [ch_Ethanol, ch_glucose]
        out.mol[[e, g]] += changes
        CO2.mol[co2] = (1-0.5111)*fermented * mw_glucose/44.0095
        CO2.phase = 'g'
        CO2.T = out.T
    
    def _optimize_N_reactors(self):
        totalcosts = self._totalcosts
        cost_old = np.inf
        self.N_reactors = 2
        self.simulate()
        cost_new = totalcosts[0]
        while cost_new < cost_old:
            self.N_reactors += 1
            self.simulate()
            cost_old = cost_new
            cost_new = totalcosts[0]
        self.N_reactors -= 1
        
    def _design(self):
        N_reactors = self.N_reactors
        if N_reactors is None:
            self._N_reactors = 2
            self._optimize_N_reactors()
            N_reactors = self._N_reactors
            self._N_reactors = None
            
        v_0 = self.outs[0].volnet
        tau = self._kwargs['tau']
        tau_0 = self.tau_0
        Design = self._results['Design']
        Design['Number of reactors'] = N_reactors
        Design.update(self._solve(v_0, tau, tau_0, N_reactors, self.V_wf))
        hx = self._cooler
        new_flow = Stream.like(self.outs[0])
        new_flow.mol /= N_reactors 
        hx.outs[0] = new_flow
        hx.ins[0] = new_flow
        hu = hx._heat_utilities[0]
        hu(self._Hnet/N_reactors, self.outs[0].T)
        hx._design(hu.duty)
        hx._cost()
        hu.duty *= N_reactors
        hu.cost *= N_reactors
        hu.flow *= N_reactors
        return Design
    
    def _cost(self):
        results = self._results
        Cost = results['Cost']
        Design = results['Design']
        N = Design['Number of reactors']
        V_i = Design['Reactor volume']
        R = V_i/self.V_0
        F_CEPCI = self.CEPCI/self.CEPCI_0
        FF = F_CEPCI * N * R**self.exp
        
        Cost['Reactors'] = FF * self.C_0
        Cost['Coolers'] = N * self._cooler._results['Cost']['Heat exchanger']
        self._power_utility(self.A_p * R * N * F_CEPCI)
        Cost['Agitators'] = FF * self.C_A
        return Cost



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 28 16:26:41 2019

@author: yoelr
"""

import lipidcane as lc
import biosteam as bs
solve_IRR = lc.lipidcane_tea.solve_IRR
grid = bs.evaluation.Grid(system=lc.lipidcane_sys,
                          metric=solve_IRR,
                          ID='IRR')

Lipid_cane = lc.Lipid_cane # The feedstock stream
def set_feed_price(feedstock_price): Lipid_cane.price = feedstock_price
grid.addparam(element=Lipid_cane,
              setter=set_feed_price,
              values=[0.030, 0.035, 0.040],
              isolated=True)

grid.addparam(Lipid_cane, lc.set_lipid_fraction, [0.02, 0.05, 0.10])

U34 = bs.find('U34')
def set_efficiency(fermentation_efficiency): U34.reset(efficiency=fermentation_efficiency)
grid.addparam(U34, set_efficiency, [0.85, 0.90, 0.95])
grid.show()


# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 10:57:12 2019

@author: yoelr
"""
from .. import Unit

class Shredder(Unit):
    _has_power_utility = True
    _has_proxystream = True
    _N_outs = 1
    
    #: Original Price (USD)
    C_0 = 2.5e6
    
    #: Original flow rate (kg/hr)
    V_0 = 500e3 
    
    #: Scaling exponent
    exp = 0.60 
    
    #: Electricity rate (kW/(kg/hr))
    electricity_rate = 0.006
    
    #: Original CEPCI 
    CEPCI_0 = 567.3
    
    def _run(self):
        self.outs[0].copylike(self.ins[0])
    
    def _cost(self):
        # Size factor
        massflow = self.ins[0].massnet
        S = massflow/self.V_0
        self._results['Cost']['Shredder'] = (self.CEPCI/self.CEPCI_0) * self.C_0*S**self.exp
        self._power_utility(massflow*self.electricity_rate)


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 15:53:14 2018

@author: yoelr
"""
import numpy as np
from .. import Unit
from fluids.pump import nema_sizes_hp
from .designtools.vacuum import _calc_MotorEfficiency, _calc_BreakEfficiency

ln = np.log
exp = np.exp

max_hp = nema_sizes_hp[-1]

# Ranges of flow rate (m3/hr) and working pressure where pumps work/
# Assume lower minimum flow rate and pressure for costing in Centrifugal single and Regenerative
pump_ranges = {'CentrifugalSingle': ((0.0000, 2271  ), (      0, 689500  )), 
               'CentrifugalDouble': ((0.2271, 2271  ), (      0, 6895000 )),
               'Gear':              ((0.2271, 681.3 ), (  68950, 13790000)),
               'MeteringPlunger':   ((0.2271, 6.813 ), (  68950, 20685000)),
               'Screw':             ((0.2271, 13.626), (  68950, 5516000 )),
               'MeteringDiaphragm': ((0.2271, 1.1355), (  68950, 5516000 )),
               'DirectactingSteam': ((0.2271, 454.2 ), (  68950, 13790000)),
               'AxialFlow':         (( 227.1, 2271  ), (  68950, 137900  ))}

# Regenerative pump (Cost not yet implemented)
regenerative_ranges = ((0.0100, 34.065), (1379000, 6895000 )) 

# Material factors
F_Mdict = {'Cast iron':       1,
           'Ductile iron':    1.15,
           'Cast steel':      1.35,
           'Bronze':          1.9,
           'Stainless steel': 2,
           'Hastelloy C':     2.95,
           'Monel':           3.3,
           'Nickel':          3.5,
           'Titanium':        9.7}

# Gear factor 
F_Tgear = {'OpenDripProof':           1,
           'EnclosedFanCooled':       1.4,
           'ExplosionProofEnclosure': 1.8}

# Centrifugal factor
F_Tcentrifugal = {'VSC3600':   1,
                  'VSC1800':   1.5,
                  'HSC3600':   1.7,
                  'HSC1800':   2,
                  '2HSC3600':  2.7,
                  '2+HSC3600': 8.9}

# Pump types
Types = ('CentrifugalSingle', 'CentrifugalDouble', 'Gear', 'MeteringPlunger', 'Default')

class Pump(Unit):
    """Create a pump that sets the pressure of the 0th output stream.

    **Parameters**

        **P:** *[float]* Pressure of output stream (Pa). If None, cost pump as increase of pressure to P_startup.
    
    **ins**
    
        [0] Input stream
        
    **outs**
    
        [1] Output stream
    
    **Examples**
    
        :doc:`Pump Example`
    
    """
    _units = {'Ideal power': 'hp',
              'Power': 'hp',
              'Head': 'ft',
              'Flow rate': 'gpm'}
    _N_ins = 1
    _N_outs = 1
    _has_power_utility = True
    _has_proxystream = True
    _kwargs = {'P': None}
    
    # Pump type
    _Type = 'Default'
    
    # Material factor
    _F_Mstr = 'Cast iron'
    _F_M = 1
    
    # Gear factor 
    _F_Tstr = 'VSC3600'
    _F_T = 1
    
    #: Pressure for costing purposes (Pa).
    P_startup = 405300 

    @property
    def Type(self):
        """Pump type"""
        return self._Type
    @Type.setter
    def Type(self, Type):
        if Type not in Types:
            Types_str = str(Types)[1:-1]
            raise ValueError('Pump type must be one of the following: ' + Types_str)
        self._Type = Type
    
    @property
    def material(self):
        """Pump material"""
        return self._material
    
    @material.setter    
    def material(self, material):
        try:
            self._F_M = F_Mdict[material]
        except KeyError:
            dummy = str(F_Mdict.keys())[11:-2]
            raise ValueError(f"Material must be one of the following: {dummy}")
        self._F_Mstr = material  
        
    def _run(self):
        P = self._kwargs['P']
        for i, o in zip(self.ins, self.outs):
            o.copylike(i)
            if P: o.P = P
    
    def _design(self):
        Design = self._results['Design']
        si = self.ins[0]
        so = self.outs[0]
        Pi = si.P
        Po = so.P
        Qi = si.volnet
        mass = si.massnet
        
        # Get type
        Type = self.Type
        if Type == 'Default':
            Type = self._default_type(Pi, Po, Qi)
        Design['Type'] = Type
        
        # Get ideal power
        if abs(Po - Pi) < 1:    
            Po = self.P_startup
        power_ideal = self._calc_PowerPressure(Pi, Po, Qi)*3.725e-7
        Design['Ideal power'] = power_ideal # hp
        
        Design['Flow rate'] = q = Qi*4.403
        if power_ideal <= max_hp:
            Design['Efficiency'] = e = self._calc_Efficiency(q, power_ideal)
            Design['Power'] = power =  self._nearest_PumpPower(power_ideal/e)
            Design['N_pumps'] = 1
            Design['Head'] = self._calc_Head(power, mass)
        else:
            raise Exception('More than 1 pump required, but not yet implemented.')
        
        self._power_utility(power/1.341)
    
    def _cost(self):
        # Parameters
        results = self._results
        Cost = results['Cost']
        Design = results['Design']
        Type = Design['Type']
        q = Design['Flow rate']
        h = Design['Head']
        p = Design['Power']
        F_M = self._F_M
        I = self.CEPCI/567
        lnp = ln(p)
        
        # Cost pump
        if 'Centrifugal' in Type:
            # Find pump factor
            F_Tdict = F_Tcentrifugal
            F_T = 1 # Assumption
            if Type == 'CentrifugalSingle':
                if 50 <= q <= 900 and 50 <= h <= 400:
                    F_T = F_Tdict['VSC3600']
                elif 50 <= q <= 3500 and 50 <= h <= 2000:
                    F_T = F_Tdict['VSC1800']
                elif 100 <= q <= 1500 and 100 <= h <= 450:
                    F_T = F_Tdict['HSC3600']
                elif 250 <= q <= 5000 and 50 <= h <= 500:
                    F_T = F_Tdict['HSC1800']
            elif Type == 'CentrifugalDouble':
                if 50 <= q <= 1100 and 300 <= h <= 1100:
                    F_T = F_Tdict['2HSC3600']
                elif 100 <= q <= 1500 and 650 <= h <= 3200:
                    F_T = F_Tdict['2+HSC3600']
            # Calculate cost
            self._S = S = self._calc_SizeFactor(q, h)
            S_new = S if S > 400 else 400
            lnS = ln(S_new)
            Cb = exp(12.1656-1.1448*lnS+0.0862*lnS**2)
            Cb *= S/S_new
            Cost['Pump'] = F_M*F_T*Cb*I
        elif Type == 'Gear':
            q_new = q if q > 50 else 50
            lnq = ln(q_new)
            Cb = exp(8.2816 - 0.2918*lnq + 0.0743*lnq**2)
            Cb *= q/q_new
            Cost['Pump'] = F_M*Cb*I
        elif Type == 'MeteringPlunger':
            Cb = exp(7.9361 + 0.26986*lnp + 0.06718*lnp**2)
            Cost['Pump'] = F_M*Cb*I
        
        # Cost electric motor
        lnp2 = lnp**2
        lnp3 = lnp2*lnp
        lnp4 = lnp3*lnp
        Cost['Motor'] = exp(5.9332 + 0.16829*lnp - 0.110056*lnp2 + 0.071413*lnp3 - 0.0063788*lnp4)*I
        return Cost
    
    @staticmethod
    def _calc_SizeFactor(q:'gal/min', h:'ft') -> 'S':
        return q*h**0.5
    
    @classmethod
    def _default_type(cls, Pi, Po, Qi):
        """Return default selection of pump type."""
        if Pi <= Po:
            for key, ranges in pump_ranges.items():
                flow_range, pressure_range = ranges
                flow_min, flow_max = flow_range
                pressure_min, pressure_max = pressure_range
                if flow_min <  Qi < flow_max and pressure_min < Pi and pressure_max > Po:
                    return key
        elif Pi > Po:
            flow_range, pressure_range = regenerative_ranges
            flow_min, flow_max = flow_range
            pressure_min, pressure_max = pressure_range
            if flow_min <  Qi < flow_max and pressure_min < Pi and pressure_max > Po:
                return 'Regenerative'
        raise Exception(f'No valid pump option for pressure, {Pi} Pa, and flow rate, {Qi} m3/hr.')
    
    @staticmethod
    def _available_PumpTypes(Pi, Po, Qi):
        """Return tuple of available pump types.
        
        **Parameters**
        
            Pi: [Stream] Input pressure (Pa)
            
            Po: [Stream] Output pressure (Pa)
            
            Qi: [Stream] Input flow rate (m^3/hr)
            
        """
        # Check which pumps work for flow rates
        pumps = [] # Pumps that work
        for key, ranges in pump_ranges.items():
            flow_range, pressure_range = ranges
            flow_min, flow_max = flow_range
            pressure_min, pressure_max = pressure_range
            if flow_min <  Qi < flow_max and pressure_min < Pi and pressure_max > Po:
                pumps.append(key)
        return pumps

    @staticmethod
    def _calc_PowerPressure(Pi, Po, Qi):
        """Return ideal power due to pressure change.
        
        **Parameters**
        
            Pi: [Stream] Input pressure
            
            Po: [Stream] Output pressure
            
            Qi: [Stream] Input flow rate
        
        """
        return Qi*(Po - Pi)
    
    @staticmethod
    def _calc_PowerFlow(Qi, Qo, Dpipe, mass=None):
        """Return ideal power due to flow rate change.
        
        **Parameters**
            
            Qi: [Stream] Input flow rate
            
            Qo: [Stream] Output flow rate
            
            Dpipe: [float] Pipe inside diameter
            
            mass: [float] Mass flow rate
        
        """
        if mass:
            A = np.pi*((Dpipe/2)**2)
            # Kinetic energy change term
            vi = Qi/A # velocity in
            vo = Qo/A # velocity out
            K_term = (mass*vo**2 - mass*vi**2)/2
        else:
            K_term = 0
        
        return K_term
    
    @staticmethod
    def _calc_Head(p:'hp', mass:'kg/hr') -> '(ft)':
        return p/mass*897806 # 897806 = 1/g * unit_conversion_factor
    
    @staticmethod
    def _nearest_PumpPower(p:'hp'):
        for power in nema_sizes_hp:
            if power >= p:
                return power
        return power
        
    @staticmethod
    def _calc_Efficiency(q:'gpm', p:'hp'):
        mup = _calc_BreakEfficiency(q)
        mum = _calc_MotorEfficiency(p/mup)
        return mup*mum
        
        